/*
 *	cellgrid.h
 *
 *	The Cell Grid class manages a collection of cells, which can be arranged
 *	in a two dimensional array.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_CELLGRID_H_
#define	_OPaC_CELLGRID_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

OP_INTERFACE (cellgrid.h);

/****************************************************************/

#ifndef _OPaC_VIEW_H_
#include <OPaC/view.h>
#endif

/****************************************************************/

/*
 *	A Cell Grid is an optimized super-view which stores
 *	visual cells into an array.
 */

class OP_PUBLIC OPCellGrid : public OPView
{
	
protected:
	OPCoord			cell_dx;					//	width of cell (if isometric)
	OPCoord			cell_dy;					//	height of cell (if isometric)
	Bool			is_isometric;				//	TRUE => each cell has the same width/height
	
	OPCoord*		cell_dx_array;				//	if not isometric, width of columns
	OPCoord*		cell_dy_array;				//	if not isometric, height of lines
	OPCoord*		cell_ox_array;				//	if not isometric, position of columns
	OPCoord*		cell_oy_array;				//	if not isometric, position of lines
	
	OPCoord			offset_x;					//	horizontal offset to first cell
	OPCoord			offset_y;					//	vertical offset to first cell
	
	OPCoord			margin;						//	margin between the frame and the cells
	OPCoord			margin_x_left;				//	margin at the left
	OPCoord			margin_x_right;				//	margin at the right
	OPCoord			margin_y_top;				//	margin at the top
	OPCoord			margin_y_bottom;			//	margin at the bottom
	
	OPCoord			min_cell_dx;				//	minimum cell width
	OPCoord			min_cell_dy;				//	minimum cell height
	OPCoord			max_cell_dx;				//	maximum cell width
	OPCoord			max_cell_dy;				//	maximum cell height
	
	Count			h_cell_count;				//	horizontal # of cell columns
	Count			v_cell_count;				//	vertical # of cell lines
	
	Bool			is_geom_ok;					//	TRUE => geometry up to date
	Bool			is_minmax_ok;				//	TRUE => min/max up to date
	Bool			draw_frame_high;			//	TRUE => draw a high rectangle
	Bool			draw_frame_low;				//	TRUE => draw a low rectangle
	
	OPControlRef*	draw_cache;					//	while drawing: pointer to current cell
	Count			draw_h_cache;				//	while drawing: horizontal position
	Count			draw_v_cache;				//	while drawing: horizontal position

	OPControlArray	array;						//	array of cell pointers
	OPControlRef	ghost_cell;					//	ghost cell (does not exist, really)
	
protected:
	OPCellGrid ();
	virtual ~ OPCellGrid ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
	virtual void ChangeGeometry (OPRectIn);				/* OVERRIDDEN */
	
	virtual void GetCellRect (Count x, Count y, OPRectOut rect) const;
	virtual void GetCellMinSize (Count x, Count y, OPCoordOut dx, OPCoordOut dy) const;
	virtual void ComputeCellMinMax ();
	virtual void ComputeMargins ();
	virtual void DiscardSizeInfo ();
	
	virtual void NotifyArraySize (Count dx, Count dy);
	virtual void NotifyAttachCell (OPControlIn cell, Count x, Count y);
	virtual void NotifyDetachCell (OPControlIn cell, Count x, Count y);
	
public:
	virtual void Delete ();								/* OVERRIDDEN */
	
	virtual Bool SetArraySize (Count dx, Count dy);
	virtual Bool SetArrayWidth (Count dx);
	virtual Bool SetArrayHeight (Count dy);
	virtual void SetMargin (OPCoord margin);
	virtual void SetIsometricMode (Bool yes);
	
	virtual void SetDrawFrameHigh (Bool yes);
	virtual void SetDrawFrameLow (Bool yes);
	
	virtual void GetArraySize (CountOut dx, CountOut dy) const;
	virtual Count ReturnArrayWidth () const;
	virtual Count ReturnArrayHeight () const;
	virtual OPCoord ReturnMargin () const;
	virtual Bool ReturnIsometricMode () const;
	
	virtual void SetCellSize (OPCoord dx, OPCoord dy);
	virtual void GetCellSize (OPCoordOut dx, OPCoordOut dy) const;
	virtual void GetCellSize (Count x, Count y, OPCoordOut dx, OPCoordOut dy) const;
	virtual void GetCellPos (Count x, Count y, OPCoordOut ox, OPCoordOut oy) const;
	
	virtual void SetCell (Count x, Count y, OPControlIn cell);
	virtual Bool GetCell (Count x, Count y, OPControlOut cell) const;
	virtual Bool FindCell (OPControlIn cell, CountOut x, CountOut y) const;
	
	virtual void UpdateCellGeometry () const;
	virtual void UpdateGridGeometry ();
	
	virtual void InsertCellInColumn (OPControlIn cell, Count x = 0);
	virtual void InsertCellInLine (OPControlIn cell, Count y = 0);
	
	//	Inherited methods :
	
	virtual void GetMinSize (OPCoordOut, OPCoordOut) const;				/* OVERRIDDEN */
	virtual void GetMaxSize (OPCoordOut, OPCoordOut) const;				/* OVERRIDDEN */
	virtual Bool Detect (OPCoord x, OPCoord y, OPViewOut);				/* OVERRIDDEN */
	virtual void Draw (OPGrailPortIn, OPRectIn, OPDrawOp = OP_DRAW_ALL);/* OVERRIDDEN */
	
	OP_DEFINE_CLASS (OPCellGrid, OPView);
};

#endif /* _OPaC_CELLGRID_H_ */

