/*
 *	sd_grailpxm.cxx
 *
 *	CAUTION: For 32-bit Windows only !
 *
 *	GRAphical Interface Layer pixel map interface class. This is
 *	heavily used for double buffering.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "sd_interface.h"

#define DeleteBitmap(hbm)      DeleteObject((HGDIOBJ)(HBITMAP)(hbm))
#define SelectBitmap(hdc, hbm) ((HBITMAP)SelectObject((hdc), (HGDIOBJ)(HBITMAP)(hbm)))

#include <OPaC/grail.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION (grail.h);
OP_IMPLEMENT_CLASS (OPGrailPixmap);

/****************************************************************/

/*
 *	Initialise the Grail Pixmap instance. This won't allocate
 *	any bitmap; do this using "Initialise" !
 */
	
OPGrailPixmap::OPGrailPixmap ()
{
	OP_METHOD (OPGrailPixmap, 0);
	
	this->pixmap = 0;
	this->dx     = 0;
	this->dy     = 0;
	this->depth  = 0;
	
	this->pixmap = OP_ALLOC_FROM_ZONE (this->ReturnMemZone (), OPaC_GrPixmap);
	
	this->pixmap->hbitmap = NULL;
	this->pixmap->hdc     = NULL;
	this->pixmap->horg_bm = NULL;
	
	this->is_cache_enabled = FALSE;
	
	this->rgb_cache[0] = 0;
	this->rgb_cache[1] = 0;
	this->rgb_cache[2] = 0;
}



/*
 *	Destroy cleanly a Pixmap : this will free an existing
 *	bitmap and associated data.
 */

OPGrailPixmap::~OPGrailPixmap ()
{
	OP_METHOD (~OPGrailPixmap, 0);
	
	this->info->hdc = NULL;
	
	if (this->pixmap) {
		
		if (this->pixmap->hdc) {
			SelectBitmap (this->pixmap->hdc, this->pixmap->horg_bm);
		}
		
		if (this->pixmap->hbitmap) {
			DeleteBitmap (this->pixmap->hbitmap);
		}
		
		if (this->pixmap->hdc) {
			DeleteDC (this->pixmap->hdc);
		}
		
		OP_FREE_FROM_ZONE (this->ReturnMemZone (), this->pixmap);
	}
}



/*
 *	Start drawing into the window. This will first draw into
 *	the buffering pixmap.
 */

void
OPGrailPixmap::BeginDraw (OPRectIn box, OPRelCoord margin)
{
	OP_METHOD (BeginDraw, 2);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPRelCoord, margin);
	OP_ASSERT (this->pixmap->hdc);
	super::BeginDraw (box, margin);
}



/*
 *	Start drawing into the pixel map. Tell the Grail Port
 *	which display context to use for drawing.
 */

void
OPGrailPixmap::BeginDraw (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OP_METHOD (BeginDraw, 4);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ASSERT (this->pixmap->hdc);
	super::BeginDraw (x1, y1, x2, y2);
}



/*
 *	End of drawing. Just propagate the message and remove
 *	the display context from the Grail Port.
 */

void
OPGrailPixmap::EndDraw ()
{
	OP_METHOD (EndDraw, 0);
	super::EndDraw ();
}



/*
 *	Initialise a new pixel map. If one already existed, just
 *	delete it and start anew. If "window" is not specified,
 *	the default screen will be used as model.
 */

void
OPGrailPixmap::Initialise (OPGrailPixmapIn window, OPCoord dx, OPCoord dy)
{
	OP_METHOD (Initialise, 3);
	OP_ARG (1, OPGrailPixmapIn, window);
	OP_ARG (2, OPCoord, dx);
	OP_ARG (3, OPCoord, dy);
	OP_ASSERT ((dx > 0) && (dy > 0));
	OP_ASSERT (this->is_drawing == FALSE);
	
	OPaC_GrPort* info = 0;
	Bool         ok   = FALSE;
	
	if (window.IsValid ()) {
		info = window->ReturnPrivatePortInfo ();
		ok   = (info && info->hdc) ? TRUE : FALSE;
	}
	
	//	Release any existing bitmap and display context information,
	//	otherwise the memory would be lost :
	
	if (this->pixmap->hdc) {
		SelectBitmap (this->pixmap->hdc, this->pixmap->horg_bm);
	}
	
	if (this->pixmap->hbitmap) {
		DeleteBitmap (this->pixmap->hbitmap);
		this->pixmap->hbitmap = NULL;
	}
	
	if (this->pixmap->hdc) {
		DeleteDC (this->pixmap->hdc);
		this->pixmap->hdc = NULL;
	}
	
	this->is_ready = FALSE;
	
	
	//	Create a new display context and then build a new bitmap...
	
	HDC      info_hdc     = (ok) ? info->hdc : GetDC (NULL);
	HPALETTE hpal_new     = this->color_man->ReturnPrivateInfo ()->hpal;
	HPALETTE hpal_old     = (hpal_new) ? SelectPalette (info_hdc, hpal_new, FALSE) : NULL;
	
	this->pixmap->hdc     = CreateCompatibleDC (NULL);
	this->pixmap->hbitmap = CreateCompatibleBitmap (info_hdc, (int)(dx), (int)(dy));
	
	if (hpal_old) {
		SelectPalette (info_hdc, hpal_old, FALSE);
	}
	
	if (ok == FALSE) ReleaseDC (NULL, info_hdc);
	
	//	If either bitmap or display context could not be allocated,
	//	this is (currently) considered to be fatal.
	
	OP_ASSERT (this->pixmap->hdc);
	OP_ASSERT (this->pixmap->hbitmap);
	
	BITMAP bm;
	GetObject (this->pixmap->hbitmap, sizeof (bm), & bm);
	this->depth = (int)(bm.bmBitsPixel * bm.bmPlanes);
	
	//	Select the bitmap and initialise it with an empty white
	//	pattern, after having activated the palette.
	
	this->pixmap->horg_bm = SelectBitmap (this->pixmap->hdc, this->pixmap->hbitmap);
	PatBlt (this->pixmap->hdc, 0, 0, (int)(dx), (int)(dy), WHITENESS);
	
	this->is_ready  = TRUE;
	this->dx        = dx;
	this->dy        = dy;
	this->info->hdc = this->pixmap->hdc;
}



/*
 *	Initialise a new pixel map with the specified depth.
 *	If one already existed, just delete it and start anew.
 */

void
OPGrailPixmap::Initialise (Count depth, OPCoord dx, OPCoord dy)
{
	OP_METHOD (Initialise, 3);
	OP_ARG (1, Count, depth);
	OP_ARG (2, OPCoord, dx);
	OP_ARG (3, OPCoord, dy);
	
	OP_ASSERT ((dx > 0) && (dy > 0));
	OP_ASSERT (this->is_drawing == FALSE);
	
	
	//	Release any existing bitmap and display context information,
	//	otherwise the memory would be lost :
	
	if (this->pixmap->hdc) {
		SelectBitmap (this->pixmap->hdc, this->pixmap->horg_bm);
	}
	
	if (this->pixmap->hbitmap) {
		DeleteBitmap (this->pixmap->hbitmap);
		this->pixmap->hbitmap = NULL;
	}
	
	if (this->pixmap->hdc) {
		DeleteDC (this->pixmap->hdc);
		this->pixmap->hdc = NULL;
	}
	
	this->is_ready = FALSE;
	
	//	Create a new display context and then build a new bitmap...
	
	HDC window_hdc    = GetDC (NULL);
	HPALETTE hpal_new = this->color_man->ReturnPrivateInfo ()->hpal;
	HPALETTE hpal_old = hpal_new ? SelectPalette (window_hdc, hpal_new, FALSE) : NULL;
	this->pixmap->hdc = CreateCompatibleDC (NULL);
	
	//	This &%*@# Win16 needs to have a pixmap compatible with the screen
	//	or it won't be happy to draw using BitBlt !
	
	if (depth == 0) {
		this->pixmap->hbitmap = CreateCompatibleBitmap (window_hdc, (int)(dx), (int)(dy));
	} else {
		this->pixmap->hbitmap = CreateBitmap ((int)(dx), (int)(dy), 1, (unsigned)(depth), NULL);
	}
	
	if (hpal_old) {
		SelectPalette (window_hdc, hpal_old, FALSE);
	}
	
	ReleaseDC (NULL, window_hdc);
	
	
	//	If either bitmap or display context could not be allocated,
	//	this is (currently) considered to be fatal.
	
	OP_ASSERT (this->pixmap->hdc);
	OP_ASSERT (this->pixmap->hbitmap);
	
	
	//	Select the bitmap and initialise it with an empty white
	//	pattern, after having activated the palette.
	
	this->pixmap->horg_bm = SelectBitmap (this->pixmap->hdc, this->pixmap->hbitmap);
	PatBlt (this->pixmap->hdc, 0, 0, (int)(dx), (int)(dy), WHITENESS);
	
	this->is_ready  = TRUE;
	this->dx        = dx;
	this->dy        = dy;
	this->depth     = depth;
	this->info->hdc = this->pixmap->hdc;
}


/*
 *	Return the RGB components of a few pixels by reading back the
 *	underlying pixmap. This implementation is system dependent.
 */

Bool
OPGrailPixmap::SystemGetRGBPixels (OPCoord x, OPCoord y, OPCoord dx, Card8* r, Card8* g, Card8* b)
{
	OP_METHOD (SystemGetRGBPixels, 6);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, Card8*, r);
	OP_ARG (5, Card8*, g);
	OP_ARG (6, Card8*, b);

	//	System dependent extraction...
	
	Card32 num  = (Card32)(dx);
	Card32 line = (Card32)(y);
	Card32 pos  = (Card32)(x);
	
	while (num--) {
		COLORREF color = GetPixel (this->pixmap->hdc, pos++, line);
		*r++ = GetRValue (color);
		*g++ = GetGValue (color);
		*b++ = GetBValue (color);
	}
	
	return TRUE;
}
