/*
 *	datamir.cxx
 *
 *	This file implements an OPaC mirror class which can be used to access
 *	data which is stored in containers, using complicated extractors.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/array.h>
#include <OPaC/datamir.h>
#include <OPaC/new.h>
#include <OPaC/values.h>

/****************************************************************/

OP_IMPLEMENTATION (datamir.h);
OP_IMPLEMENT_CLASS (OPDataMirror);

/****************************************************************/

struct OPDataMirror::DataMirrorCache
{
	Bool			is_ok;					//	TRUE => cache is valid (may be used)
	Bool			is_enabled;				//	TRUE => cache is enabled
	OPObjectRef		object;					//	cached object
};


/*
 *	Initialise a data mirror. By default, the cache will
 *	be enabled.
 */

OPDataMirror::OPDataMirror ()
{
	OP_METHOD (OPDataMirror, 0);
	
	this->x_defined = FALSE;
	this->index     = 0;
	
	//	The cache is handled as a separate piece of memory, otherwise
	//	it would not be possible to have a const GetObject method...
	
	this->cache = NEW_FROM_ZONE (this->ReturnMemZone ()) DataMirrorCache;

	this->cache->is_ok      = FALSE;
	this->cache->is_enabled = TRUE;
}



/*
 *	Clean-up the data mirror. Just release the memory used
 *	by the cache.
 */

OPDataMirror::~OPDataMirror ()
{
	OP_METHOD (~OPDataMirror, 0);
	DELETE_FROM_ZONE (this->ReturnMemZone (), this->cache);
}



/*
 *	Store the data mirror. This preserves everything in a
 *	dynamic way.
 */

void
OPDataMirror::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);

	storage << this->container;
	storage << this->x_defined;
	storage << (this->x_defined ? this->x_args : OPObjectNull);
	storage << this->index;
	storage << this->cache->is_enabled;
	storage << OPStorageSync::mark;
}



/*
 *	Restore the data mirror.
 */

void
OPDataMirror::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->container;
	storage >> this->x_defined;
	storage >> this->x_args;
	storage >> this->index;
	storage >> this->cache->is_enabled;
	storage >> OPStorageSync::mark;

	this->cache->is_ok  = FALSE;
	this->cache->object = 0;
}



/*
 *	Prepare for deletion. Don't delete the object which
 *	is being referred to by this instance.
 */

void
OPDataMirror::Delete ()
{
	OP_METHOD (Delete, 0);

	super::Delete ();

	this->container     = 0;
	this->x_args        = 0;
	this->cache->is_ok  = FALSE;
	this->cache->object = 0;
}



/*
 *	Setup the container. No check is performed.
 */

void
OPDataMirror::SetContainer (OPGenericContainerIn source)
{
	OP_METHOD (SetContainer, 1);
	OP_ARG (1, OPGenericContainerIn, source);
	
	this->container     = source;
	this->cache->is_ok  = FALSE;
	this->cache->object = 0;
}



/*
 *	Define the index to use for any further accesses using
 *	either GetObject or one of the QueryData/DefineData
 *	methods.
 *
 *	No check is performed.
 */

void
OPDataMirror::SetIndex (Count index)
{
	OP_METHOD (SetIndex, 1);
	OP_ARG (1, Count, index);
	
	if (this->index == index) {
		return;
	}
	
	this->index         = index;
	this->cache->is_ok  = FALSE;
	this->cache->object = 0;
}



/*
 *	Flush the cache. The next access to the object will
 *	refresh the cache.
 */

void
OPDataMirror::FlushCache () const
{
	OP_METHOD (FlushCache, 0);

	this->cache->is_ok  = FALSE;
	this->cache->object = 0;
}



/*
 *	Enable or disable the cache. If the cache is enabled,
 *	the object will be only fetched once; if the contents
 *	of the container change, this won't get noticed.
 */

void
OPDataMirror::EnableCache (Bool state) const
{
	OP_METHOD (EnableCache, 1);
	OP_ARG (1, Bool, state);
	
	if (state == this->cache->is_enabled) {
		return;
	}
	
	this->FlushCache ();
	
	this->cache->is_enabled = state;
}



/*
 *	Return the object which is specified by the container,
 *	the extraction and the index. If this is not possible,
 *	then return FALSE.
 */

Bool
OPDataMirror::GetObject (OPObjectOut object) const
{
	OP_METHOD (GetObject, 1);
	OP_ARG (1, OPObjectOut, object);
	
	if (this->cache->is_ok) {
		object = this->cache->object;
		return TRUE;
	}
	
	if ( (this->container.IsValid ())
	  && ((this->x_defined == FALSE) || (this->container->SpecifyExtraction (x_args)))
      && (this->container->Get (this->index, object))
      && (object.IsValid ()) ) {
		
		if (this->cache->is_enabled) {
			this->cache->is_ok  = TRUE;
			this->cache->object = object;
		}
		
		return TRUE;
	}

	return FALSE;
}



/*
 *	Return total count of container which this instance
 *	is mirroring.
 */

Count
OPDataMirror::TotalCount () const
{
	OP_METHOD (TotalCount, 0);
	
	return this->container.IsValid ()
	     ? this->container->TotalCount ()
	     : 0;
}



/*
 *	Return the nth object of the container which this
 *	instance is mirroring.
 *
 *	This does not alter the index specified using the
 *	SetIndex method. It does not rely on the cache
 *	either.
 */

Bool
OPDataMirror::Get (Count index, OPObjectOut object)
{
	OP_METHOD (Get, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPObjectOut, object);
	
	if ( (this->container.IsValid ())
	  && ((this->x_defined == FALSE) || (this->container->SpecifyExtraction (this->x_args)))
      && (this->container->Get (index, object))
      && (object.IsValid ()) ) {
		return TRUE;
	}

	return FALSE;
}



/*
 *	Specify both the extraction name and the arguments
 *	associated to it. See SetExtractionName/Args.
 *
 *	Always return TRUE.
 */

Bool
OPDataMirror::SpecifyExtraction (OPArrayIn x)
{
	OP_METHOD (SpecifyExtraction, 1);
	OP_ARG (1, OPArrayIn, x);
	
	this->x_defined     = TRUE;
	this->x_args        = x;
	this->cache->is_ok  = FALSE;
	this->cache->object = 0;
	
	return TRUE;
}



/*
 *	Some containers (lists, etc.) don't need any extraction
 *	and thus there is a method to remove the previously
 *	selected extraction.
 */

void
OPDataMirror::SpecifyNoExtraction ()
{
	OP_METHOD (SpecifyNoExtraction, 0);
	
	this->x_defined = FALSE;
	this->x_args    = 0;
}



/*
 *	Count the public data of the mirrored object. Return 0
 *	if no object can be found.
 */

Count
OPDataMirror::CountPublicData () const
{
	OP_METHOD (CountPublicData, 0);

	OPObjectRef object;

	if (this->GetObject (object)) {
		return object->CountPublicData ();
	}

	return 0;
}



/*
 *	Return the nth public data name of the mirrored object
 *	currently defined by the container/extraction/index.
 */

Bool
OPDataMirror::GetPublicData (Count index, OPAtomOut name) const
{
	OP_METHOD (GetPublicData, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPAtomOut, name);
	
	OPObjectRef object;

	if (this->GetObject (object)) {
		return object->GetPublicData (index, name);
	}

	return FALSE;
}



/*
 *	Return the nth public data informations (name, type
 *	and shortcut) of the mirrored object currently defined
 *	by the container/extraction/index.
 */

Bool
OPDataMirror::GetPublicData (Count index,
							 OPAtomOut name,
							 OPTypeDefOut type) const
{
	OP_METHOD (GetPublicData, 3);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPAtomOut, name);
	OP_ARG (3, OPTypeDefOut, type);
	
	OPObjectRef object;
	
	if (this->GetObject (object)) {
		return object->GetPublicData (index, name, type);
	}

	return FALSE;
}

Bool
OPDataMirror::GetPublicData (Count index, OPAtomOut name, OPStringOut desc,
							 OPTypeDefOut type,
							 BoolOut read_ok, BoolOut write_ok) const
{
	OP_METHOD (GetPublicData, 6);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPAtomOut, name);
	OP_ARG (3, OPStringOut, desc);
	OP_ARG (4, OPTypeDefOut, type);
	OP_ARG (5, BoolOut, read_ok);
	OP_ARG (6, BoolOut, write_ok);
	
	OPObjectRef object;
	
	if (this->GetObject (object)) {
		return object->GetPublicData (index, name, desc, type, read_ok, write_ok);
	}

	return FALSE;
}

Bool
OPDataMirror::GetPublicDataNotifier (OPAtomIn name, Count nth_notifier,
									 OPAtomOut notifier) const
{
	OP_METHOD (GetPublicDataNotifier, 3);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, Count, nth_notifier);
	OP_ARG (3, OPAtomOut, notifier);
	
	OPObjectRef object;
	
	if (this->GetObject (object)) {
		return object->GetPublicDataNotifier (name, nth_notifier, notifier);
	}
	
	return FALSE;
}

Bool
OPDataMirror::GetPublicDataEx (Count index,
							   Count nth_bit,
							   const OPaC_BitsetDataDef*& def,
							   Card32Out num_bits) const
{
	OP_METHOD (GetPublicDataEx, 4);
	OP_ARG (1, Count, index);
	OP_ARG (2, Count, nth_bit);
	OP_ARG (3, const OPaC_BitsetDataDef*&, def);
	OP_ARG (4, Card32Out, num_bits);
	
	OPObjectRef object;
	
	if (this->GetObject (object)) {
		return object->GetPublicDataEx (index, nth_bit, def, num_bits);
	}
	
	return FALSE;
}


/*
 *	Return the data associated to the specified name, for
 *	the currently mirrored object.
 */
	
Bool
OPDataMirror::QueryData (OPAtomIn name, OPObjectOut data) const
{
	OP_METHOD (QueryData, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPObjectOut, data);
	
	if (name.IsInvalid ()) {
		return this->GetObject (data);
	}
	
	OPObjectRef object;
	
	if (this->GetObject (object)) {
		return object->QueryData (name, data);
	}
	
	return FALSE;
}

Bool
OPDataMirror::QueryData (OPAtomIn name, OPValueOut data) const
{
	OP_METHOD (QueryData, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPValueOut, data);
	
	if (name.IsInvalid ()) {
		
		OPObjectRef object;
		
		if (this->GetObject (object)) {
			data->SetObject (object);
			return TRUE;
		}
		
		return FALSE;
	}
	
	OPObjectRef object;
	
	if (this->GetObject (object)) {
		return object->QueryData (name, data);
	}
	
	return FALSE;
}

Bool
OPDataMirror::QueryData (Count index, OPObjectOut data) const
{
	OP_METHOD (QueryData, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPObjectOut, data);
	
	OPObjectRef object;
	
	if (this->GetObject (object)) {
		return object->QueryData (index, data);
	}
	
	return FALSE;
}

Bool
OPDataMirror::QueryData (Count index, OPValueOut data) const
{
	OP_METHOD (QueryData, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPValueOut, data);
	
	OPObjectRef object;
	
	if (this->GetObject (object)) {
		return object->QueryData (index, data);
	}
	
	return FALSE;
}



/*
 *	Define the data associated to the specified name, for
 *	the currently mirrored object.
 */

Bool
OPDataMirror::DefineData (OPAtomIn name, OPObjectIn data)
{
	OP_METHOD (DefineData, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPObjectIn, data);
	
	OPObjectRef object;
	
	if (this->GetObject (object)) {
		return object->DefineData (name, data);
	}
	
	return FALSE;
}

Bool
OPDataMirror::DefineData (OPAtomIn name, OPValueIn data)
{
	OP_METHOD (DefineData, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPValueIn, data);
	
	OPObjectRef object;
	
	if (this->GetObject (object)) {
		return object->DefineData (name, data);
	}
	
	return FALSE;
}

Bool
OPDataMirror::DefineData (Count index, OPObjectIn data)
{
	OP_METHOD (DefineData, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPObjectIn, data);
	
	OPObjectRef object;
	
	if (this->GetObject (object)) {
		return object->DefineData (index, data);
	}
	
	return FALSE;
}

Bool
OPDataMirror::DefineData (Count index, OPValueIn data)
{
	OP_METHOD (DefineData, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPValueIn, data);
	
	OPObjectRef object;
	
	if (this->GetObject (object)) {
		return object->DefineData (index, data);
	}
	
	return FALSE;
}

