/*
 *	dataview.cxx
 *
 *	Data View class implementation.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/dataview.h>
#include <OPaC/datamir.h>
#include <OPaC/new.h>
#include <OPaC/pubmacro.h>
#include <OPaC/texttag.h>
#include <OPaC/values.h>

/****************************************************************/

OP_IMPLEMENTATION (dataview.h);
OP_IMPLEMENT_CLASS (OPDataView);

/****************************************************************/

/*
 *	Initialise an empty data view. By default, the text is
 *	not considered to be volatile (cache enabled).
 */

OPDataView::OPDataView ()
{
	OP_METHOD (OPDataView, 0);

	this->state |= (OP_VIEW_OK_DETECT			//	may be detected (click)
				  | OP_VIEW_OK_MAIN				//	may become main responder
				  | OP_VIEW_IS_ALIVE);			//	may become active
	
	this->format   = OPStringNull;
	this->text     = OPStringNull;
	this->text_tag = OPTextTagBaseNull;
}



/*
 *	Clean-up the data view.
 */

OPDataView::~OPDataView ()
{
	OP_METHOD (~OPDataView, 0);
}



/*
 *	Store the data view : the links needed to display the
 *	data will be maintained.
 */

void
OPDataView::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->text_tag << this->format;
	storage << OPStorageSync::mark;
}



/*
 *	Restore the data view. Also restore the links to the
 *	data and the other informations which are needed.
 */

void
OPDataView::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->text_tag >> this->format;
	storage >> OPStorageSync::mark;
}



/*
 *	Make the proper string representation.
 */

void
OPDataView::BuildTextFromData ()
{
	OP_METHOD (BuildTextFromData, 0);
	
	//	Ask the "data" to fill the text field with a readable representation
	//	of the published data. If FALSE is returned, an empty string will be
	//	used instead.
	
	OPValueRef value = this->values[0];
	
	if (value.IsValid ()) {
		value->ConvertToString (this->text, this->format);
	} else {
		this->text = "";
	}
}

void
OPDataView::DoUpdate ()
{
	OP_METHOD (DoUpdate, 0);
	
	this->BuildTextFromData ();
	
	if (this->text_tag.IsInvalid ()) {
		this->InitTextTag (OP_NEW (OPTextTag));
	}
	
	this->text_tag->SetText (this->text);
}


void
OPDataView::ChangeGeometry (OPRectIn box)
{
	OP_METHOD (ChangeGeometry, 1);
	OP_ARG (1, OPRectIn, box);
	
	super::ChangeGeometry (box);
	
	if (this->text_tag.IsValid ()) {
		this->text_tag->SetBox (box);
	}
}


/*
 *	Prepare the data view for its destruction. Just remove
 *	any links to the outer world.
 */

void
OPDataView::Delete ()
{
	OP_METHOD (Delete, 0);

	this->format   = OPStringNull;
	this->text     = OPStringNull;
	this->text_tag = OPTextTagBaseNull;
	
	super::Delete ();
}


/*
 *	Specify a special text tag used to display the text.
 *	This can be used to replace the standard text tag.
 */

Bool
OPDataView::InitTextTag (OPTextTagBaseIn special_text_tag)
{
	OP_METHOD (InitTextTag, 1);
	OP_ARG (1, OPTextTagBaseIn, special_text_tag);
	
	if (this->text_tag == special_text_tag) {
		return TRUE;
	}
	
	if (this->text_tag.IsValid ()) {
		this->text_tag->SetParent (OPViewNull);
		this->text_tag->Delete ();
	}
	
	this->text_tag = special_text_tag;
	
	if (this->text_tag.IsValid ()) {
		this->text_tag->SetParent (this);
		this->text_tag->SetBox (this->box);
	}
	
	return TRUE;
}


/*
 *	Define the format of the data conversion.
 */

Bool
OPDataView::InitValueFormat (OPStringIn format)
{
	OP_METHOD (InitValueFormat, 1);
	OP_ARG (1, OPStringIn, format);
	
	this->format = format;
	this->FlushCache ();
	
	return TRUE;
}



/*
 *	Return the text's minimal size.
 */

void
OPDataView::GetMinSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	if (this->text_tag.IsInvalid ()) {
		OPDataView* that = (OPDataView*)(this);		/* UGLY const-away cast */
		that->UpdateCache ();
	}
	
	if (this->text_tag.IsValid ()) {
		this->text_tag->GetMinSize (dx, dy);
	} else {
		super::GetMinSize (dx, dy);
	}
}
