/*
 *	filebrowser.cxx
 *
 *	The File Browser helps browsing directories.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/


#include <OPaC/database.h>
#include <OPaC/icon.h>
#include <OPaC/filebrowser.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/string.h>
#include <OPaC/values.h>

#include <OPaC/pubmacro.h>

#include <OPaC/sd/filedir.h>

#include <string.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPFileBrowser);

OP_PUBLISH_FUNCTION (OPFileBrowser, "Provider", ReturnProvider, "Provider", OPObject, OPObject);

/****************************************************************/


OPFileBrowser::OPFileBrowser ()
{
	OP_METHOD (OPFileBrowser, 0);
}

OPFileBrowser::~OPFileBrowser ()
{
	OP_METHOD (~OPFileBrowser, 0);
}

void
OPFileBrowser::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->path << this->db << this->db_model << this->provider;
	storage << OPStorageSync::mark;
}

void
OPFileBrowser::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->path >> this->db >> this->db_model >> this->provider;
	storage >> OPStorageSync::mark;
}


void
OPFileBrowser::Initialise (OPStringIn path)
{
	OP_METHOD (Initialise, 1);
	OP_ARG (1, OPStringIn, path);
	
	this->path = path;
}

void
OPFileBrowser::Update ()
{
	OP_METHOD (Update, 0);
	
	static OPAtomRef atom_index = "index";
	static OPAtomRef atom_name  = "name";
	static OPAtomRef atom_data  = "data";
	static OPAtomRef atom_text  = "text";
	static OPAtomRef atom_large_icon = "large icon";
	static OPAtomRef atom_small_icon = "small icon";
	static OPAtomRef atom_mark_icon  = "mark icon";
	
	if (this->db.IsInvalid ()) {
		
		this->db       = OP_NEW (OPDataBase);
		this->db_model = OP_NEW (OPEntityModel);
		
		this->db_model->Initialise (this->db, "ListModel", 7);
		
		this->db_model->AddAttribute (atom_index);
		this->db_model->AddAttribute (atom_name);
		this->db_model->AddAttribute (atom_data);
		this->db_model->AddAttribute (atom_text);
		this->db_model->AddAttribute (atom_large_icon);
		this->db_model->AddAttribute (atom_small_icon);
		this->db_model->AddAttribute (atom_mark_icon);
		
	} else {
		this->db->RemoveAll ();
	}
	
	if (this->provider.IsInvalid ()) {
		
		this->provider = OP_NEW (OPDBProvider);
		
		this->provider->Initialise (this->db);
		
		this->provider->SetDBNameForSorting (atom_index);
		this->provider->SetDBNameForObject (atom_data);
		this->provider->SetDBNameForText (atom_text);
		this->provider->SetDBNameForLargeIcon (atom_large_icon);
		this->provider->SetDBNameForSmallIcon (atom_small_icon);
		this->provider->SetDBNameForMarkIcon (atom_mark_icon);
	}
	
	OPaC_FileDir* dir = OPaC_FileDir::New ();
	
	if (dir->Open (this->path, "*")) {
		
		OPaC_FileInfo info;
		
		while (dir->FindNext (info)) {
			
			OPEntityRef entity = OP_NEW (OPEntity);
			OPStringRef name   = info.name->ReturnString ();
			OPFileBrowserRef browser;
			
			if (info.is_directory) {
				browser = OP_NEW (OPFileBrowser);
				OPStringRef new_path = this->path->ReturnString ();
				new_path->AppendString ("/");
				new_path->AppendString (name);
				browser->Initialise (new_path);
				browser->SetAttribute ("DefaultBrowserProvider", browser);
			}
			
			OPStringRef index;
			
			if (info.is_directory) {
				index = " ";
				index->AppendString (name);
			} else {
				index = name;
			}
			
			entity->Initialise (this->db_model);
			entity->SetAttribute (1, index);
			entity->SetAttribute (2, name);
			entity->SetAttribute (3, browser);
			entity->SetAttribute (4, name);
			entity->SetAttribute (5, (OPValueRef)(Int32)(OP_ICON_NONE));
			entity->SetAttribute (6, (OPValueRef)(Int32)(info.is_directory ? OP_ICON_TREE_DIR : OP_ICON_TREE_FILE));
			entity->SetAttribute (7, (OPValueRef)(Int32)(info.is_directory ? OP_ICON_SUBMENU_NORMAL : OP_ICON_NONE));
			
			this->db->Insert (entity);
		}
	}
	
	OPaC_FileDir::Delete (dir);
}


const OPDBProviderRef&
OPFileBrowser::ReturnProvider () const
{
	OP_METHOD (ReturnProvider, 0);
	
	if (this->provider.IsInvalid ()) {
		OPFileBrowser* that = const_cast<OPFileBrowser*> (this);
		that->Update ();
	}
	
	return this->provider;
}

