/*
 *	grailpath.cxx
 *
 *	GRAphical Interface Layer, path building. The path building
 *	and manipulation classes are used to handle filled regions.
 *
 *	(C) Copyright 1997, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/grail.h>
#include <OPaC/grailpath.h>
#include <OPaC/new.h>

OP_IMPLEMENTATION (grailpath.h);
OP_IMPLEMENT_CLASS (OPGrailPath);

/****************************************************************/

OPGrailPath::OPGrailPath ()
{
	this->InitPathToZero ();
}

OPGrailPath::~OPGrailPath ()
{
	this->ClearPath ();
}


/*
 *	Free the memory allocated for the path description.
 */

void
OPGrailPath::ClearPath ()
{
	OP_METHOD (ClearPath, 0);
	
	Path* path = this->path;
	Path* next = 0;
	
	while (path) {
		next = (path->more) ? (path->more) : (path->next);
		OP_FREE (path);
		path = next;
	}

	this->InitPathToZero ();
}


/*
 *	Set the fields to zero.
 */

void
OPGrailPath::InitPathToZero ()
{
	OP_METHOD (InitPathToZero, 0);
	
	this->ox = this->oy = 0;
	this->cx = this->cy = 0;
	this->px = this->py = 0;
	
	this->path = 0;
	this->current_path = 0;
	this->current_path_part = 0;
	
	this->is_path_open  = FALSE;
	this->is_current_ok = FALSE;
	this->is_optimised  = TRUE;
	this->is_segment_in_path = FALSE;
	
	this->bounds.MakeEmpty ();
}


/*
 *	Add a new segment to the current path. If no path is currently
 *	open, just allocate a new piece of path. If the current piece
 *	of path is full, link it to a new one.
 *
 *	A path is said to be optimised when the segments do not have
 *	gaps between them.
 */

void
OPGrailPath::AddPathSegment (OPCoord x, OPCoord y)
{
	OP_METHOD (AddPathSegment, 2);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	
	this->is_optimised = FALSE;
	this->bounds.InsertPoint (x, y);
	
	if (this->is_path_open == FALSE) {
		
		Path* newpath = OP_ALLOC (Path);
		Path* update  = this->current_path;
		
		newpath->next = 0;
		newpath->more = 0;
		newpath->segments = 0;
		
		if (update == 0) {
			this->path = newpath;
		}
		
		while (update) {
			update->next = newpath;
			update = update->more;
		}
		
		this->current_path      = newpath;
		this->current_path_part = newpath;
		
		this->is_path_open = TRUE;
	}
	
	Path* path = this->current_path_part;
	
	if (path->segments == MAX_PATH_SEG) {
		
		Path* newpath = OP_ALLOC (Path);
		
		newpath->next = 0;
		newpath->more = 0;
		newpath->segments = 0;
		
		path->more = newpath;
		
		this->current_path_part = newpath;
	}
	
	if ( (this->is_segment_in_path)
	  && ( (this->px != this->cx)
	    || (this->py != this->cy) ) ) {
		this->is_optimised = FALSE;
	}
	
	path->seg[path->segments].p1.x = this->cx;
	path->seg[path->segments].p1.y = this->cy;
	path->seg[path->segments].p2.x = this->cx = this->px = x;
	path->seg[path->segments].p2.y = this->cy = this->py = y;
	
	path->segments++;
	this->is_segment_in_path = TRUE;
}

void
OPGrailPath::OptimisePath ()
{
	if (this->is_optimised) return;
	
	this->is_optimised = TRUE;
}


const OPRect&
OPGrailPath::GetBoundingBox () const
{
	return this->bounds;
}

Bool
OPGrailPath::GetCurrentPoint (OPCoordOut x, OPCoordOut y) const
{
	if (this->is_current_ok) {
		x = this->cx;
		y = this->cy;
		return TRUE;
	}
	
	return FALSE;
}


Bool
OPGrailPath::NewPath ()
{
	this->ClearPath ();
	return TRUE;
}

Bool
OPGrailPath::ClosePath ()
{
	if (this->is_path_open) {
		this->LineTo (this->ox, this->oy);
		this->is_path_open = FALSE;
		this->is_segment_in_path = FALSE;
	}
	
	return TRUE;
}

Bool
OPGrailPath::MoveTo (OPCoord x, OPCoord y)
{
	this->cx = x;
	this->cy = y;
	this->is_current_ok = TRUE;
	return TRUE;
}

Bool
OPGrailPath::LineTo (OPCoord x, OPCoord y)
{
	if (this->is_current_ok) {
		this->AddPathSegment (x, y);
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPGrailPath::CurveTo ()
{
	if (this->is_current_ok) {
		
		//	Do the job...
		
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPGrailPath::Arc (OPCoord x, OPCoord y, OPCoord r, OPCoord a1, OPCoord a2,
				  Bool counterclockwise)
{
	if (this->is_current_ok) {
		
		//	Do the job...
		
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPGrailPath::ArcTangent (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	if (this->is_current_ok) {
		
		//	Do the job...
		
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPGrailPath::Stroke (OPGrailPortIn port, OPCoord x, OPCoord y)
{
	if (this->is_optimised == FALSE) {
		this->OptimisePath ();
	}
	
	Path* path = this->path;

	while (path) {
		
		const Segment* seg = path->seg;
		Count          num = path->segments;
		
		while (num--) {
			port->DrawHairLineTo (seg->p1.x, seg->p1.y, seg->p2.x, seg->p2.y);
			seg++;
		}
		
		path = (path->more) ? (path->more) : (path->next);
	}
	
	return TRUE;
}


Bool
OPGrailPath::Fill (OPGrailPortIn port, OPCoord x, OPCoord y)
{
	if (this->is_optimised == FALSE) {
		this->OptimisePath ();
	}
	
	return TRUE;
}

