/*
 *	grailpxm.cxx
 *
 *	GRAphical Interface Layer pixel map interface class. This is
 *	heavily used for double buffering. System independent part.
 *
 *	(C) Copyright 1994-1997, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/grail.h>
#include <OPaC/new.h>

/****************************************************************/

/*
 *	Save the pixmap to an archive. This has not been
 *	implemented yet.
 */

void
OPGrailPixmap::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	super::StoreInstance (storage);
	
	//	NIY
}


/*
 *	Restore a pixmap from an archive.
 */

void
OPGrailPixmap::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	super::RestoreInstance (storage);
	
	//	NIY
}



/*
 *	Return the RGB compontents of a few pixels by either reading back
 *	the cache (if any) or by using the system dependent implementation.
 *	This is the only public method available for reading back pixels.
 */

Bool
OPGrailPixmap::GetRGBPixels (OPCoord x, OPCoord y, OPCoord dx, Card8* r, Card8* g, Card8* b)
{
	OP_METHOD (GetRGBPixels, 6);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, Card8*, r);
	OP_ARG (5, Card8*, g);
	OP_ARG (6, Card8*, b);

	if (this->is_cache_enabled) {
		
		Card32 num    = (Card32)(dx);
		Card32 line   = (Card32)(y);
		Card32 pos    = 0;
		Card32 pos_x  = (Card32)(x);
		Card8* rgb[3] = { 0, 0, 0 };
		Card8* dst[3] = { r, g, b };
		
		//	If the line is not yet stored in the cache, read it by using the
		//	system dependent extraction.
		
		if (this->AccessCacheRGBLine (line, rgb[0], rgb[1], rgb[2]) == FALSE) {
			this->SystemGetRGBPixels (0, y, this->dx, rgb[0], rgb[1], rgb[2]);
		}
		
		//	Copy the RGB components required by the caller. An horizontal
		//	offset can be specified.
		
		while (num--) {
			for (int i = 0; i < 3; i++) {
				dst[i][pos] = rgb[i][pos_x];
			}
			pos++;
			pos_x++;
		}
		
		return TRUE;
	}
	
	//	If the cache is disabled, use the system dependent extraction...
	
	return this->SystemGetRGBPixels (x, y, dx, r, g, b);
}

void
OPGrailPixmap::EnableCache (Bool enable)
{
	OP_METHOD (EnableCache, 1);
	OP_ARG (1, Bool, enable);
	
	this->is_cache_enabled = enable;
	this->FlushCache ();
}

void
OPGrailPixmap::FlushCache ()
{
	OP_METHOD (FlushCache, 0);
	
	//	Free all the memory which had been allocated to the cache (if any).
	
	for (int i = 0; i < 3; i++) {
		if (this->rgb_cache[i]) {
			for (int y = 0; y < this->dy; y++) {
				if (this->rgb_cache[i][y]) {
					OP_FREE (this->rgb_cache[i][y]);
				}
			}
			
			OP_FREE (this->rgb_cache[i]);
			
			this->rgb_cache[i] = 0;
		}
	}
}


/*
 *	Return the pointers into the R, G and B components of a cached line.
 *	This can be used to fill the cache directly. Return TRUE if the line
 *	had already been allocated before.
 */

Bool
OPGrailPixmap::AccessCacheRGBLine (int line, Card8*& r, Card8*& g, Card8*& b)
{
	OP_METHOD (AccessCacheRGBLine, 4);
	OP_ARG (1, int, line);
	OP_ARG (2, const Card8*&, r);
	OP_ARG (3, const Card8*&, g);
	OP_ARG (4, const Card8*&, b);
	
	Bool    ready  = TRUE;
	Card8** rgb[3] = { &r, &g, &b };
	
	for (int i = 0; i < 3; i++) {
		if (this->rgb_cache[i] == 0) {
			this->rgb_cache[i] = OP_ALLOC_ARRAY (Card8*, (int)this->dy);
			for (int y = 0; y < (int)this->dy; y++) {
				this->rgb_cache[i][y] = 0;
			}
		}
		if (this->rgb_cache[i][line] == 0) {
			ready = FALSE;
			this->rgb_cache[i][line] = OP_ALLOC_ARRAY (Card8, (int)this->dx);
		}
		*(rgb[i]) = this->rgb_cache[i][line];
	}
	
	return ready;
}


void
OPGrailPixmap::CacheRGBLine (int line, const Card8* r, const Card8* g, const Card8* b)
{
	OP_METHOD (CacheRGBLine, 4);
	OP_ARG (1, int, line);
	OP_ARG (2, const Card8*, r);
	OP_ARG (3, const Card8*, g);
	OP_ARG (4, const Card8*, b);
	
	const Card8* source_rgb[3] = { r, g, b };
	Card8* destination_rgb[3]  = { 0, 0, 0 };
	
	//	Find the pointers into the RGB line cache. Update the contents of the cache
	//	with the source RGB values.
	
	this->AccessCacheRGBLine (line, destination_rgb[0], destination_rgb[1], destination_rgb[2]);
	
	for (int i = 0; i < 3; i++) {
		for (int x = 0; x < (int)(this->dx); x++) {
			destination_rgb[i][x] = source_rgb[i][x];
		}
	}
}

Bool
OPGrailPixmap::FindRGBLine (int line, Card8* r, Card8* g, Card8* b)
{
	OP_METHOD (FindRGBLine, 4);
	OP_ARG (1, int, line);
	OP_ARG (2, Card8*, r);
	OP_ARG (3, Card8*, g);
	OP_ARG (4, Card8*, b);
	
	Card8* rgb[3] = { r, g, b };
	
	for (int i = 0; i < 3; i++) {
		if (this->rgb_cache[i] == 0) return FALSE;
		if (this->rgb_cache[i][line] == 0) return FALSE;
		
		for (int x = 0; x < (int)this->dx; x++) {
			rgb[i][x] = this->rgb_cache[i][line][x];
		}
	}
	
	return TRUE;
}
