/*
 *	list.cxx
 *
 *	The OPaC List class provides a sorted list in which objects
 *	can be stored. Their position depends on their contents.
 *
 *	(C) Copyright 1994-1997, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/object.h>
#include <OPaC/container.h>
#include <OPaC/listit.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPList);

/****************************************************************/


/*
 *	Create an empty list, which will by default be sorted.
 */

OPList::OPList ()
{
	OP_METHOD (OPList, 0);
	
	this->is_sorted      = TRUE;
	this->sort_on_insert = TRUE;
	
	this->first = 0;
	this->last  = 0;
}


/*
 *	Destroy the list. Every element maintained in the list
 *	will be removed. If no longer referenced, these will
 *	automagically be deleted.
 */

OPList::~OPList ()
{
	OP_METHOD (~OPList, 0);
	
	if (this->num_element) {
		this->RemoveAll ();
	}
}



/*
 *	Save a list to an archive.
 */

void
OPList::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->is_sorted << this->sort_on_insert;
	storage << OPStorageSync::mark;
	
	Element* elem = this->first;
	
	while (elem) {
		storage << elem->data;
		elem = elem->next;
	}
	
	storage << OPStorageSync::mark;
}



/*
 *	Restore a list from an archive.
 */

void
OPList::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->is_sorted >> this->sort_on_insert;
	storage >> OPStorageSync::mark;
	
	Count num = this->num_element;
	
	while (num--) {
		
		Element* elem = NEW_FROM_ZONE (this->ReturnMemZone ()) Element;
		
		elem->next = 0;
		elem->prev = this->last;
		elem->data = 0;
		
		(elem->prev ? elem->prev->next : this->first) = elem;
		(elem->next ? elem->next->prev : this->last)  = elem;
		
		storage >> elem->data;
	}
	
	storage >> OPStorageSync::mark;
}



/*
 *	Prepare the list for its destruction.
 */

void
OPList::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->RemoveAll ();
	super::Delete ();
}


OPIteratorRef
OPList::NewIterator ()
{
	OPListIteratorRef iterator = OP_NEW (OPListIterator);
	
	this->InsertIterator (iterator);
	this->iterators->Insert (iterator);
	return iterator;
}



/*
 *	Insert an object into the list. If the list is defined
 *	as sorted, then insert at the right place.
 *
 *	Return TRUE if successful. Invalid objects will not be
 *	accepted.
 */

Bool
OPList::Insert (OPObjectIn value)
{
	OP_METHOD (Insert, 1);
	OP_ARG (1, OPObjectIn, value);
	
	if (value.IsValid () == FALSE) {
		return FALSE;
	}
	
	Element* elem = NEW_FROM_ZONE (this->ReturnMemZone ()) Element;
	
	if (this->sort_on_insert) {
		
		//	If the list is sorted, we will do a sorted insertion.
		
		Element* find = this->first;
		
		//	Find the element which will become the successor in the sorted list...
		//	This is quite ineffective - should try to use splay or Red-Black trees.
		
		while (find) {
			
			if (find->data->CompareWith (value) == OP_COMP_THIS_AFTER) {
				break;
			}
			
			find = find->next;
		}
		
		elem->next = (find) ? find : 0;
		elem->prev = (find) ? find->prev : this->last;
		
	} else {
		
		//	If the list is no longer sorted, then just insert at the end of the
		//	list...
		
		elem->next = 0;
		elem->prev = this->last;

		this->is_sorted = FALSE;
	}
	
	elem->data = value;
	(elem->prev ? elem->prev->next : this->first) = elem;
	(elem->next ? elem->next->prev : this->last)  = elem;
	this->num_element++;
	
	if (this->iterators.IsValid ()) {
		Count index = 0;
		OPListIteratorRef iter;
		
		while (this->iterators->Get (index++, iter)) {
			if (iter.IsValid ()) {
				iter->NotifyInsert (elem);
			}
		}
	}
	
	return TRUE;
}



/*
 *	Remove the nth object which is equal to the specified
 *	one. Equality means same pointers.
 *
 *	Return TRUE if successful.
 */

Bool
OPList::Remove (OPObjectIn value, Count nth)
{
	OP_METHOD (Remove, 2);
	OP_ARG (1, OPObjectIn, value);
	OP_ARG (2, Count, nth);
	
	Element* elem = this->first;
	
	while (elem) {
		
		if ( (elem->data == value)
		  && (nth-- == 0) ) {
			
			if (this->iterators.IsValid ()) {
				Count index = 0;
				OPListIteratorRef iter;
				
				while (this->iterators->Get (index++, iter)) {
					if (iter.IsValid ()) {
						iter->NotifyRemove (elem);
					}
				}
			}
			
			(elem->next ? elem->next->prev : this->last)  = elem->prev;
			(elem->prev ? elem->prev->next : this->first) = elem->next;
			this->num_element--;
			
			elem->data = 0;
			
			DELETE (elem);
			
			return TRUE;
		}
		
		elem = elem->next;
	}
	
	return FALSE;
}



/*
 *	Remove an object defined by its position in the list.
 *	Return TRUE if successful.
 */

Bool
OPList::Remove (Count pos)
{
	OP_METHOD (Remove, 1);
	OP_ARG (1, Count, pos);
	
	if (pos >= this->num_element) {
		return FALSE;
	}
	
	Element* elem  = this->first;
	
	while (pos--) {
		elem = elem->next;
	}
	
	(elem->next ? elem->next->prev : this->last)  = elem->prev;
	(elem->prev ? elem->prev->next : this->first) = elem->next;
	this->num_element--;
	OPObjectRef& value = elem->data;
	value = 0;
	DELETE (elem);
	
	return TRUE;
}



/*
 *	Remove every object contained in the list.
 */

void
OPList::RemoveAll ()
{
	OP_METHOD (RemoveAll, 0);
	
	while (this->first) {
		Element*     elem  = this->first;
		OPObjectRef& value = elem->data;
		this->first = elem->next;
		value = 0;
		DELETE (elem);
	}
	
	this->first       = 0;
	this->last        = 0;
	this->is_sorted   = TRUE;
	this->num_element = 0;
}



/*
 *	Find the object at the specified position in the list.
 *	This will return TRUE if successful.
 */

Bool
OPList::Get (Count pos, OPObjectOut object)
{
	OP_METHOD (Get, 2);
	OP_ARG (1, Count, pos);
	OP_ARG (2, OPObjectOut, object);
	
	if (pos >= this->num_element) {
		return FALSE;
	}
	
	Element* elem = this->first;
	
	while (pos--) {
		elem = elem->next;
	}
	
	object = elem->data;
	
	return TRUE;
}



/* 
 *	Look for some similar object in the list (take the nth,
 *	if nth == 0 take the first one). Comparison is based
 *	on the "CompareWith" method. Return TRUE if successful.
 */

Bool
OPList::GetSimilar (OPObjectIn value, OPObjectOut object, Count nth)
{
	OP_METHOD (GetSimilar, 3);
	OP_ARG (1, OPObjectIn, value);
	OP_ARG (2, OPObjectOut, object);
	OP_ARG (3, Count, nth);
	
	Element* elem = this->first;
	
	while (elem) {
		
		if ( (elem->data->CompareWith (value) == OP_COMP_THIS_SAME)
		  && (nth-- == 0) ) {
			
			object = elem->data;
			return TRUE;
		}
		
		elem = elem->next;
	}
	
	return FALSE;
}



/*
 *	Return TRUE if the object can be found in the list.
 *	The comparison is based on the pointers.
 */

Bool
OPList::LookFor (OPObjectIn value)
{
	OP_METHOD (LookFor, 1);
	OP_ARG (1, OPObjectIn, value);
	
	Element* elem = this->first;
	
	while (elem) {
		
		if (elem->data == value) {
			return TRUE;
		}
		
		elem = elem->next;
	}
	
	return FALSE;
}


/*
 *	Return the state of the list. If it has been sorted once
 *	and if the insertions have also taken into account the
 *	sorting, then the list still is sorted.
 */

Bool
OPList::IsSorted () const
{
	return this->is_sorted;
}


/*
 *	Define insertion mode. Default is automatic sorting while
 *	inserting elements.
 */

Bool
OPList::SetInsertMode (Bool sort)
{
	this->sort_on_insert = sort;
	return TRUE;
}

Bool
OPList::GetInsertMode () const
{
	return this->sort_on_insert;
}
