/*
 *	objectwell.cxx
 *
 *	The Object Well class implements a drop target for objects...
 *	
 *	(C) Copyright 1997-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/app.h>
#include <OPaC/button.h>
#include <OPaC/container.h>
#include <OPaC/grail.h>
#include <OPaC/icon.h>
#include <OPaC/lookman.h>
#include <OPaC/message.h>
#include <OPaC/new.h>
#include <OPaC/objectwell.h>
#include <OPaC/textstyle.h>

#include <OPaC/pubmacro.h>
#include <OPaC/targetmacro.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPObjectWell);

OP_TARGET_ACTOR (OPObjectWell, Update);
OP_TARGET_ACTOR (OPObjectWell, KillObject);

OP_TARGET_CLASS (OPObjectWell);
OP_TARGET_ARG (OPDataReportRef, reporter);
OP_TARGET_CALL (AddReporter (reporter));
OP_TARGET_END ();

OP_PUBLISH_MODIFIER (OPObjectWell, "Object", GetObject, SetObject, "Object", OPObject, OPObject);
OP_PUBLISH_SETVALUE (OPObjectWell, "ObjectEditor", SetObjectEditor, "Object editor", OPObject, OPView);

/****************************************************************/

OPCoord		OPObjectWell::well_arrow_dx = 0;
OPCoord		OPObjectWell::well_dx       = 0;
OPCoord		OPObjectWell::well_dy       = 0;
OPCoord		OPObjectWell::well_border   = 0;

/****************************************************************/

OPObjectWell::OPObjectWell ()
{
	OP_METHOD (OPObjectWell, 0);
	
//	this->state &= ~ OP_VIEW_OK_UI_EDIT;
	this->state |=   OP_VIEW_OK_DROP;
	this->shape &= ~ OP_CTRL_RELIEF;
	this->shape |=   OP_CTRL_NOFRAME;
	
	this->has_edit_support  = TRUE;
	this->has_right_arrow   = FALSE;
	this->has_direct_editor = FALSE;
	
	this->editor_width = 150;
	
	OPLookManRef look = OPLookMan::Current ();
	
	look->GetIconSize (OP_ICON_RIGHT_ARROW, OPObjectWell::well_arrow_dx, OPObjectWell::well_dy);
	look->GetIconSize (OP_ICON_CUBE_NORMAL, OPObjectWell::well_dx, OPObjectWell::well_dy);
	
	OPObjectWell::well_border = look->Return3DEffectWidth ();
}

OPObjectWell::~OPObjectWell ()
{
}

void
OPObjectWell::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->CleanupWell ();
	this->object = OPObjectNull;
	
	super::Delete ();
}

void
OPObjectWell::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->object;
	storage << OPStorageSync::mark;
}

void
OPObjectWell::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->object;
	storage >> OPStorageSync::mark;
}


/*
 *	Clean-up the internal objects set up with "SetupWellForObject"
 *	so that an initial (empty) state is recovered.
 */

void
OPObjectWell::CleanupWell ()
{
	OP_METHOD (CleanupWell, 0);
	
	if (this->editor.IsValid ()) {
		this->editor->SetParent (OPViewNull);
		this->editor->Delete ();
		this->editor = 0;
	}
	
	if (this->kill_icon.IsValid ()) {
		this->kill_icon->SetParent (OPViewNull);
		this->kill_icon->Delete ();
		this->kill_icon = 0;
	}
	
	if (this->kill_button.IsValid ()) {
		this->kill_button->SetParent (OPViewNull);
		this->kill_button->Delete ();
		this->kill_button = 0;
	}
	
 	if (this->reporters.IsValid ()) {
		
		OPObjectRef reporter;
		
		while (this->reporters->Get (0, reporter)) {
			reporter->Delete ();
			this->reporters->Remove (0);
		}
		
		this->reporters->Delete ();
		this->reporters = 0;
	}
	
	this->has_right_arrow   = FALSE;
	this->has_direct_editor = FALSE;
}


/*
 *	Set up the well for the newly defined object. This will
 *	create the editor, if needed.
 */

void
OPObjectWell::SetupWellForObject ()
{
	OP_METHOD (SetupWellForObject, 0);
	
	this->CleanupWell ();
	
	OPCoord dx;
	OPCoord dy;
	
	this->GetMinSize (dx, dy);
	this->SetSize (dx, dy);
	
	static OPAtomRef sel_attr = "EmbeddedObjectForConnect";
	static OPAtomRef uie_attr = "EmbeddedObjectForUIEdit";
	
	this->RemoveAttribute (sel_attr);
	this->RemoveAttribute (uie_attr);
	
	if ( (this->has_edit_support)
	  && (this->object.IsValid ()) ) {
		
		this->SetAttribute (sel_attr, this->object);
		this->SetAttribute (uie_attr, this->object);
		this->has_right_arrow = TRUE;
		
		OPObjectRef builder;
		
		OPApp::CurrentApp ()->GetBuilder (builder);
		
		if (builder.IsValid ()) {
			
			//	If there is a valid builder attached to the containing window,
			//	ask it to provide the simple object editor needed to edit the
			//	current object.
			
			OPMessageRef doEditSimpleObject = OP_NEW_FROM_ZONE (OPaC_temp_zone, OPMessage);
			OPAtomRef editor_name = "ObjectEditor";
			doEditSimpleObject->SetMethod ("EditSimpleObject");
			
			builder <- doEditSimpleObject << this << this->object << editor_name;
			
			//	The builder will define the "editor" field if a new editor has been
			//	created. Otherwise, the editor will still be invalid.
			
			this->has_direct_editor = this->editor.IsValid ();
		}
		
		//	Find out about the co-ordinates of the objects which will be
		//	added to the right of the well.
		
		OPCoord x = this->box.X1 ();
		OPCoord y = this->box.Y1 ();
		
		x += OPObjectWell::well_dx + 2*OPObjectWell::well_border;
		x += this->has_right_arrow ? OPObjectWell::well_arrow_dx : 0;
		
		//	If a direct editor has been set up, incorporate it into the
		//	well, so that it will be available to the user.
		
		if (this->has_direct_editor) {
			
			OPCoord dx;
			OPCoord dy;
			
			this->editor->GetSize (dx, dy);
			this->editor->SetSize (this->editor_width, dy);
			this->editor->SetParent (this);
			this->editor->SetOrigin (x, y);
			this->editor->SetFlagOkUIEdit (FALSE);
		}
		
		y = this->box.Y2 () - 14;
		
		this->kill_icon   = OP_NEW (OPIcon);
		this->kill_button = OP_NEW (OPButton);
		
		this->kill_icon->Initialise (OP_ICON_MINCLOSE_NORMAL);
		this->kill_button->SetShapeStyle (this->kill_button->ReturnShapeStyle () & ~OP_CTRL_AUTOACTIVE | OP_CTRL_NOFRAME);
		this->kill_button->SetParent (this);
		this->kill_button->SetOrigin (x, y);
		this->kill_button->SetSize (14, 14);
		this->kill_button->SetContents (this->kill_icon);
		this->kill_button->AttachObserver (this, "OnClick", "KillObject");
		this->kill_button->SetFlagOkUIEdit (FALSE);
	}
}

/****************************************************************/

/*
 *	Add a reporter to the list of reporters used temporarily
 *	by the well. When deleting the well, we will also delete
 *	the reporters.
 */

void
OPObjectWell::AddReporter (OPDataReportIn reporter)
{
	OP_METHOD (AddReporter, 1);
	OP_ARG (1, OPDataReportIn, reporter);
	
	if (this->reporters.IsInvalid ()) {
		this->reporters = OP_NEW (OPHashTable);
	}
	
	this->reporters->Insert (reporter);
}


/*
 *	Define the object editor.
 */

Bool
OPObjectWell::SetObjectEditor (OPViewIn view)
{
	OP_METHOD (SetObjectEditor, 1);
	OP_ARG (1, OPViewIn, view);
	
	this->editor = view;
	
	return TRUE;
}

/****************************************************************/

/*
 *	Return the well's minimal size. This takes into account the
 *	editor...
 */

void
OPObjectWell::GetMinSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	dx  = OPObjectWell::well_dx;
	dy  = OPObjectWell::well_dy;
	dx += 2*OPObjectWell::well_border;
	dy += 2*OPObjectWell::well_border;
	dx += this->has_edit_support ? OPObjectWell::well_arrow_dx : 0;
	dx += this->has_edit_support ? this->editor_width : 0;
}




Bool
OPObjectWell::SetObject (OPObjectIn object)
{
	OP_METHOD (SetObject, 1);
	OP_ARG (1, OPObjectIn, object);
	
	this->object = object;
	this->SetupWellForObject ();
	this->Display ();
	
	return TRUE;
}

Bool
OPObjectWell::GetObject (OPObjectOut object) const
{
	OP_METHOD (GetObject, 1);
	OP_ARG (1, OPObjectOut, object);
	
	object = this->object;
	
	return TRUE;
}


/*
 *	By default, simply accept the drop if the view is said to
 *	accept drops, but don't do anything.
 *
 *	Override this method to do something meaningful.
 */

Bool
OPObjectWell::DragAndDrop (OPRectIn box, OPObjectIn initiator, OPDragDropProtocol ddp)
{
	OP_METHOD (DragAndDrop, 3);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPObjectIn, initiator);
	OP_ARG (3, OPDragDropProtocol, ddp);
	
	if (ddp == OP_DDP_DRAGnDROP_COMMIT) {
		
		super::DragAndDrop (box, initiator, OP_DDP_DRAGnDROP_CANCEL);
		
		OPObjectRef object;
		initiator->QueryData ("DragObject", object);
		
		this->AdjustObject (object);
		this->SetObject (object);
		this->Report ("ObjectChange", OP_REPORT_NEW);
		this->Display ();
		
		return TRUE;
	}
	
	return super::DragAndDrop (box, initiator, ddp);
}

void
OPObjectWell::AdjustObject (OPObjectIn object)
{
	OP_METHOD (AdjustObject, 0);
	OP_ARG (1, OPObjectIn, object);
	
	if ( (object.IsValid ())
	  && (object->IsKindOf (OP_RUNTIME (OPView))) ) {
		
		//	The object is a view... Adjust its size if it happens to be too
		//	small in its current configuration.
		
		OPCoord dx_min  = 0;
		OPCoord dy_min  = 0;
		OPCoord dx_real = 0;
		OPCoord dy_real = 0;
		
		OPViewRef view = object;
		OPViewRef parent;
		
		view->GetParent (parent);
		view->GetMinSize (dx_min, dy_min);
		view->GetSize (dx_real, dy_real);
		
		if ( (dx_real != dx_min)
		  || (dy_real != dy_min) ) {
			view->SetSize (dx_min, dy_min);
		}
	}
}


void
OPObjectWell::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	const OPAdornerRef& adorner = port->ReturnAdorner ();
	const OPLookManRef& lookman = OPLookMan::Current ();
	
	super::Draw (port, rect, what);
	
	OPCoord x   = this->box.X1 ();
	OPCoord y   = this->box.Y1 ();
	OPRect  box = this->box;
	OPCoord dx  = OPObjectWell::well_dx + 2*OPObjectWell::well_border;
	OPCoord dy  = OPObjectWell::well_dy + 2*OPObjectWell::well_border;
	
	box.SetSize (dx, dy);
	
	adorner->DrawRectHigh (box);
	adorner->DrawIcon (x+OPObjectWell::well_border,
					   y+OPObjectWell::well_border, OP_ICON_CUBE_NORMAL);
	
	if (has_right_arrow) {
		
		adorner->DrawIcon (x+dx, y+OPObjectWell::well_border, OP_ICON_RIGHT_ARROW);
		
		dx += OPObjectWell::well_arrow_dx;
		dx += 20;
		
		box.SetOrigin (x+dx, y);
		box.SetSize (this->box.DX ()-dx, this->box.DY ());
		
		OPStringRef text = "class ";
		text->AppendString (this->object->ReturnRuntime ()->ReturnClassName ());
		Count length = text->CountCharacters ();
		
		adorner->SetForeColor (OP_COLOR_VIEW_NORMAL);
		port->ShowText (box, text, 0, length, OP_TEXT_HPOS_LEFT | OP_TEXT_VPOS_BOTTOM);
	}
}


Bool
OPObjectWell::Update ()
{
	OP_METHOD (Update, 0);
	
	this->AdjustObject (this->object);
	
	if ( (this->object.IsValid ())
	  && (this->object->IsKindOf (OP_RUNTIME (OPView))) ) {
		OPViewRef view = this->object;
		view->Display ();
	}
	
	return TRUE;
}

Bool
OPObjectWell::KillObject ()
{
	OP_METHOD (KillObject, 0);
	
	this->SetObject (OPObjectNull);
	this->Report ("ObjectChange", OP_REPORT_NEW);
	this->Display ();
	
	return TRUE;
}
	
