/*
 *	rbtree.cxx
 *
 *	Red-Black Tree Class implementation. This class provides a fast
 *	way to insert and retrieve sorted information from a well balanced
 *	tree.
 *
 *	Theory on red-black trees can be found in DDJ April 1992 and in
 *	Sedgewick's "Algorithms in C++".
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/rbtree.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION (rbtree.h);
OP_IMPLEMENT_CLASS (OPRBTree);

/****************************************************************/

static const Card8 NODE_NONE  = 0;
static const Card8 NODE_RIGHT = 1;
static const Card8 NODE_LEFT  = 2;
static const Card8 NODE_BOTH  = 3;

/****************************************************************/

/*
 *	Build the base red-black tree structure with the dummy
 *	head node and the dummy zero node...
 */

OPRBTree::OPRBTree ()
{
	OP_METHOD (OPRBTree, 0);
	
	this->head = & this->head_node;
	this->zero = & this->zero_node;
	
	this->head_node.left     = this->zero;
	this->head_node.right    = this->zero;
	this->head_node.is_red   = FALSE;
	this->head_node.is_valid = FALSE;
	this->head_node.key      = 0;
	this->head_node.value    = 0;
	
	this->zero_node.left     = this->zero;
	this->zero_node.right    = this->zero;
	this->zero_node.is_red   = FALSE;
	this->zero_node.is_valid = FALSE;
	this->zero_node.key      = 0;
	this->zero_node.value    = 0;
}



/*
 *	Destroy the red-black tree. This will need to free the
 *	nodes which have been stored in the tree, but not the
 *	associated data.
 */

OPRBTree::~OPRBTree ()
{
	OP_METHOD (~OPRBTree, 0);
	
	if (this->head->right != this->zero) {
		this->FreeSubTree (this->head->right);
	}
}




/*
 *	Store the red-black tree.
 */

void
OPRBTree::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	if (this->head->right == this->zero) {
		storage << NODE_NONE;
	} else {
		storage << NODE_RIGHT;
		this->StoreSubTree (this->head->right, storage);
	}
	
	storage << OPStorageSync::mark;
}



/*
 *	Restore the red-black tree.
 */

void
OPRBTree::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	Card8 status = 0;
	
	storage >> status;
	
	if (status == NODE_NONE) {
		this->head->right = this->zero;
		this->head->left  = this->zero;
	} else {
		this->head->right = this->AllocNode ();
		this->head->left  = this->zero;
		this->RestoreSubTree (this->head->right, storage);
	}
	
	storage >> OPStorageSync::mark;
}



/*
 *	Allocate some memory to hold a node. This will return
 *	an initialised node. It might be lots faster...
 */

OPRBTree::Node*
OPRBTree::AllocNode ()
{
	OP_METHOD (AllocNode, 0);
	
	Node* node = NEW_FROM_ZONE (this->ReturnMemZone ()) Node;
	
	node->left     = this->zero;
	node->right    = this->zero;
	node->is_red   = FALSE;
	node->is_valid = TRUE;
	node->key      = 0;
	node->value    = 0;
	
	return node;
}



/*
 *	Free some node and destroy the reference to the data,
 *	if there is still one.
 */

void
OPRBTree::FreeNode (Node* node)
{
	OP_METHOD (FreeNode, 1);
	OP_ARG (1, Node*, node);
	OP_ASSERT (node->is_valid);
	
	node->value = 0;
	
	DELETE (node);
}


/*
 *	Free the nodes, including the specified one. This is
 *	heavily recursive.
 */

void
OPRBTree::FreeSubTree (Node* node)
{
	OP_METHOD (FreeSubTree, 1);
	OP_ARG (1, Node*, node);
	
	if (node != this->zero) {
		
		//	In order to avoid using too much stack, the recursion is done only
		//	for each second tree level.
		
		if (node->right != this->zero) {
			
			if (node->right->right != this->zero) {
				this->FreeSubTree (node->right->right);
			}
			
			if (node->right->left != this->zero) {
				this->FreeSubTree (node->right->left);
			}
			
			this->FreeNode (node->right);
		}
		
		if (node->left != this->zero) {
			
			if (node->left->right != this->zero) {
				this->FreeSubTree (node->left->right);
			}
			
			if (node->left->left != this->zero) {
				this->FreeSubTree (node->left->left);
			}
			
			this->FreeNode (node->left);
		}
		
		this->FreeNode (node);
	}
}



/*
 *	Store a sub-tree, including the specified node. This
 *	is quite recursive...
 */

void
OPRBTree::StoreSubTree (Node* node, OPArchiveIn archive)
{
	OP_METHOD (StoreSubTree, 2);
	OP_ARG (1, Node*, node);
	OP_ARG (2, OPArchiveIn, archive);
	
	archive << node->is_red << node->key << node->value;
	
	if (node->right == this->zero) {
		
		if (node->left == this->zero) {
			
			//	Neither right nor left sub-tree. Nothing more to store !
			
			archive << NODE_NONE;
			
		} else {
			
			//	Only a left node, just store the left sub-tree.
			
			archive << NODE_LEFT;
			this->StoreSubTree (node->left, archive);
		}
		
	} else {
		
		if (node->left == this->zero) {
			
			//	Only a right node, just store the right sub-tree.
			
			archive << NODE_RIGHT;
			
		} else {
			
			//	Both right and left nodes, store both sub-trees.
			
			archive << NODE_BOTH;
			this->StoreSubTree (node->left, archive);
		}
		
		this->StoreSubTree (node->right, archive);
	}
}



/*
 *	Restore a sub-tree, including the specified node. This
 *	is quite recursive...
 */

void
OPRBTree::RestoreSubTree (Node* node, OPArchiveIn archive)
{
	OP_METHOD (RestoreSubTree, 2);
	OP_ARG (1, Node*, node);
	OP_ARG (2, OPArchiveIn, archive);
	
	Card8 status = 0;
	
	archive >> node->is_red >> node->key >> node->value;
	archive >> status;
	
	switch (status) {
		
		case NODE_NONE:
			node->right = this->zero;
			node->left  = this->zero;
			break;
		
		case NODE_LEFT:
			node->left  = this->AllocNode ();
			node->right = this->zero;
			this->RestoreSubTree (node->left, archive);
			break;
		
		case NODE_RIGHT:	
			node->left  = this->zero;
			node->right = this->AllocNode ();
			this->RestoreSubTree (node->right, archive);
			break;
		
		case NODE_BOTH:
			node->left  = this->AllocNode ();
			node->right = this->AllocNode ();
			this->RestoreSubTree (node->left, archive);
			this->RestoreSubTree (node->right, archive);
			break;
		
		default:
			OP_ABORT ("Bad node");
			break;
	}
}



/*
 *	Insert a node into the red-black tree; the position of the node
 *	is given by its key...
 */

void
OPRBTree::InsertNode (Node* n)
{
	OP_METHOD (Insert, 1);
	OP_ARG (1, Node*, n);
	OP_ASSERT (n);
	
	n->right  = this->zero;
	n->left   = this->zero;
	n->is_red = FALSE;
	
	Card32  v = n->key;					//	key used for the insertion
	Node*   x;							//	current node
	Node*   p;							//	parent node
	Node*   g;							//	grandparent node
	Node*   gg;							//	great-grandparent node
	
#define	SPLIT()												\
	do {													\
		x->is_red = TRUE;									\
		x->left->is_red = x->right->is_red = FALSE;			\
															\
		if (p->is_red) {									\
			g->is_red = TRUE;								\
			if ((v < g->key) != (v < p->key))				\
				ROTATE (g, p);								\
			ROTATE (gg, x);									\
			x->is_red = FALSE;								\
		}													\
	} while (0)

#define	ROTATE(y,z)											\
	do {													\
		Node* c = (v < y->key) ? y->left : y->right;		\
		Node* gc;											\
															\
		if (v < c->key) {									\
			gc        = c->left;							\
			c->left   = gc->right;							\
			gc->right = c;									\
		} else {											\
			gc        = c->right;							\
			c->right  = gc->left;							\
			gc->left  = c;									\
		}													\
															\
		(v < y->key) ? y->left : y->right = gc;				\
		z = gc;												\
	} while (0)


	p = g = gg = x = this->head;
	
	while (x != this->zero) {
		
		gg = g;										//	grandparent => great-grandparent
		g  = p;										//	parent => grandparent
		p  = x;										//	current => parent
		
		x  = (v < x->key) ? x->left : x->right;		//	move down the tree
		
		if (x->left->is_red && x->right->is_red)	//	split 4-entry nodes on the way down
			SPLIT ();
	}
	
	x = n;
	
	(v < p->key) ? p->left : p->right = x;			//	link current with the previous one
	SPLIT ();										//	the link to the parent is red !
	
	head->right->is_red = FALSE;					//	but.. the dummy head never becomes red
	
#undef SPLIT
#undef ROTATE
}



/*
 *	Remove a node from the tree and return a pointer to it.
 *	If no node matching can be found, return a null pointer.
 */

OPRBTree::Node*
OPRBTree::RemoveNode (Card32 v)
{
	OP_METHOD (Remove, 1);
	OP_ARG (1, Card32, v);
	
	Node* c;
	Node* p;
	Node* x;
	Node* t;
	
	this->zero->key = v;					//	needed by algorithm, provides automatic stop
	
	p = this->head;
	x = this->head->right;
	
	while (v != x->key) {
		p = x;
		x = (v < x->key) ? x->left : x->right;
	}
	
	t = x;
	
	if (t->right == this->zero) {
		
		x = t->left;
		x->is_red = FALSE;
		
	} else if (t->right->left == this->zero) {
		
		x = t->right;
		x->left = t->left;
		
		if (t->is_red == FALSE) x->is_red = FALSE;
		else if (x->right->is_red == FALSE) x->is_red = TRUE;
		
	} else {
		
		c = x->right;
		
		while (c->left->left != this->zero)
			c = c->left;
		
		x = c->left;
		
		x->is_red = t->is_red;
		if (c->is_red) x->right->is_red = FALSE;
		
		c->left  = x->right;
		x->left  = t->left;
		x->right = t->right;
	}
	
	(v < p->key) ? p->left : p->right = x;
	
	return (t == this->zero) ? 0 : t;
}



/*
 *	Find a node in the tree which key matches exactly the
 *	specified key. Return a null pointer if no matching node
 *	can be found.
 */

OPRBTree::Node*
OPRBTree::FindEQ (Card32 v)
{
	OP_METHOD (FindEQ, 1);
	OP_ARG (1, Card32, v);
	
	Node*         x = this->head->right;
	this->zero->key = v;
	
	while (v != x->key)
		x = (v < x->key) ? x->left : x->right;
	
	return (x == this->zero) ? 0 : x;
}



/*
 *	Find a node in the tree which key is larger or equal to the
 *	specified key. If there is more than one node, select the
 *	nearest one. Return a null pointer if no matching node is
 *	found.
 */

OPRBTree::Node*
OPRBTree::FindGE (Card32 v)
{
	OP_METHOD (FindGE, 1);
	OP_ARG (1, Card32, v);
	
	Node*         x = this->head->right;
	Node*         l = this->head;
	this->zero->key = v;
	
	while (v != x->key) {
		
		if (v < x->key) {
			l  = x;
			x  = x->left;
		} else {
			x  = x->right;
		}
	}
	
	return (x == this->zero) ? ( (l == this->head) ? 0 : l ) : x;
}



/*
 *	Find a node in the tree which key is smaller or equal to the
 *	specified key. If there is more than one node, select the
 *	nearest one. Return a null pointer if no matching node is
 *	found.
 */

OPRBTree::Node*
OPRBTree::FindLE (Card32 v)
{
	OP_METHOD (FindLE, 1);
	OP_ARG (1, Card32, v);
	
	Node*         x = this->head->right;
	Node*         l = this->head;
	this->zero->key = v;

	while (v != x->key) {
		
		if (v > x->key) {
			l  = x;
			x  = x->left;
		} else {
			x  = x->right;
		}
	}
	
	return (x == this->zero) ? ( (l == this->head) ? 0 : l ) : x;
}



/*
 *	Find the node with the smallest key in the whole tree.
 *	If there is no node, return a null pointer.
 */

OPRBTree::Node*
OPRBTree::FindMin () const
{
	OP_METHOD (FindMin, 0);
	
	Node* x = this->head->right;
	
	while (x->left != this->zero) {
		x = x->left;
	}
	
	return (x == this->zero) ? 0 : x;
}



/*
 *	Find the node with the largest key in the whole tree.
 *	If there is no node, return a null pointer.
 */

OPRBTree::Node*
OPRBTree::FindMax () const
{
	OP_METHOD (FindMax, 0);
	
	Node* x = this->head->right;
	
	while (x->right != this->zero) {
		x = x->right;
	}
	
	return (x == this->zero) ? 0 : x;
}


/****************************************************************/


/*
 *	Delete the red-black tree... Prepare the tree for its
 *	destruction.
 */

void
OPRBTree::Delete ()
{
	OP_METHOD (Delete, 0);
	
	if (this->head->right != this->zero) {
		this->FreeSubTree (this->head->right);
	}
	
	super::Delete ();
}


/*
 *	Insert an object into the red-black tree, based on the
 *	specified index.
 */

void
OPRBTree::Insert (Card32 index, OPObjectIn value)
{
	OP_METHOD (Insert, 2);
	OP_ARG (1, Card32, index);
	OP_ARG (2, OPObjectIn, value);
	
	Node* node = this->AllocNode ();
	
	node->key   = index;
	node->value = value;
	
	this->InsertNode (node);
}



/*
 *	Remove the node specified by the index. If there are
 *	more than one node with the same index, there is no
 *	way to know which one will be removed...
 */

Bool
OPRBTree::Remove (Card32 index)
{
	OP_METHOD (Remove, 1);
	OP_ARG (1, Card32, index);
	
	Node* node = this->RemoveNode (index);
	
	if (node) {
		this->FreeNode (node);
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Remove the node specified by the index and return
 *	the value which was stored in it.
 */

Bool
OPRBTree::Remove (Card32 index, OPObjectOut value)
{
	OP_METHOD (Remove, 2);
	OP_ARG (1, Card32, index);
	OP_ARG (2, OPObjectOut, value);
	
	Node* node = this->RemoveNode (index);
	
	if (node) {
		value = node->value;
		this->FreeNode (node);
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Look for a node with a matching index.
 */

Bool
OPRBTree::Find (Card32 index)
{
	OP_METHOD (Find, 1);
	OP_ARG (1, Card32, index);
	return this->FindEQ (index) ? TRUE : FALSE;
}



/*
 *	Look for a node with a matching index and return the
 *	associated value.
 */

Bool
OPRBTree::Find (Card32 index, OPObjectOut value)
{
	OP_METHOD (Find, 2);
	OP_ARG (1, Card32, index);
	OP_ARG (2, OPObjectOut, value);
	
	Node* node = this->FindEQ (index);
	
	if (node) {
		value = node->value;
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Look for the nearest (previous) node with a matching
 *	index.
 */

Bool
OPRBTree::FindNearestPrev (Card32 index)
{
	OP_METHOD (Find, 1);
	OP_ARG (1, Card32, index);
	return this->FindLE (index) ? TRUE : FALSE;
}



/*
 *	Look for the nearest (previous) node with a matching
 *	index and return the associated value.
 */

Bool
OPRBTree::FindNearestPrev (Card32 index, OPObjectOut value)
{
	OP_METHOD (Find, 2);
	OP_ARG (1, Card32, index);
	OP_ARG (2, OPObjectOut, value);
	
	Node* node = this->FindLE (index);
	
	if (node) {
		value = node->value;
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Look for the nearest (following) node with a matching
 *	index.
 */

Bool
OPRBTree::FindNearestNext (Card32 index)
{
	OP_METHOD (Find, 1);
	OP_ARG (1, Card32, index);
	return this->FindGE (index) ? TRUE : FALSE;
}



/*
 *	Look for the nearest (following) node with a matching
 *	index and return the associated value.
 */

Bool
OPRBTree::FindNearestNext (Card32 index, OPObjectOut value)
{
	OP_METHOD (Find, 2);
	OP_ARG (1, Card32, index);
	OP_ARG (2, OPObjectOut, value);
	
	Node* node = this->FindGE (index);
	
	if (node) {
		value = node->value;
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Return index of first element (if any).
 */

Bool
OPRBTree::FindFirst (Card32& index) const
{
	OP_METHOD (FindFirst, 1);
	OP_ARG (1, Card32&, index);
	
	Node* node = this->FindMin ();
	
	if (node) {
		index = node->key;
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Return index of last element (if any).
 */

Bool
OPRBTree::FindLast (Card32& index) const
{
	OP_METHOD (FindLast, 1);
	OP_ARG (1, Card32&, index);
	
	Node* node = this->FindMax ();
	
	if (node) {
		index = node->key;
		return TRUE;
	}
	
	return FALSE;
}
