/*
 *	scrview.cxx
 *
 *	The Scroll View is used to hold a set of views, which may be scrolled
 *	around on the screen.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/except.h>
#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/scroller.h>
#include <OPaC/scrview.h>
#include <OPaC/targetmacro.h>
#include <OPaC/values.h>
#include <OPaC/view.h>
#include <OPaC/window.h>

/****************************************************************/

OP_IMPLEMENTATION (scrview.h);
OP_IMPLEMENT_CLASS (OPScrollView);

/****************************************************************/

#define	SCROLL_HOR	((Card32)(1))
#define	SCROLL_VER	((Card32)(2))

/****************************************************************/

/*
 *	Export the observer which is used to signal scrolls
 *	to the ScrollView :
 */

OP_TARGET_OBSERVER (OPScrollView, Scroll);

/****************************************************************/

Bool      OPScrollView::name_init_ok = FALSE;
OPAtomRef OPScrollView::name_value;
OPAtomRef OPScrollView::name_ratio;
OPAtomRef OPScrollView::do_display;
OPAtomRef OPScrollView::do_scroll;

/****************************************************************/

/*
 *	Construct the Scroll View. This just resets every
 *	field.
 */

OPScrollView::OPScrollView ()
{
	OP_METHOD (OPScrollView, 0);
	
	this->h_scroller = 0;
	this->v_scroller = 0;
	
	this->x_offset_ratio = 0.0;
	this->y_offset_ratio = 0.0;
	
	this->x_offset = 0;
	this->y_offset = 0;
	
	this->width  = 0;
	this->height = 0;
	
	if (OPScrollView::name_init_ok == FALSE) {
		OPScrollView::name_value   = "VisibleValue";
		OPScrollView::name_ratio   = "VisibleRatio";
		OPScrollView::do_display   = "Display";
		OPScrollView::do_scroll    = "Scroll";
		OPScrollView::name_init_ok = TRUE;
	}
}



/*
 *	Destroying a Scroll View...
 */

OPScrollView::~OPScrollView ()
{
	OP_METHOD (~OPScrollView, 0);
}



/*
 *	Store a Scroll View with its links to the scrollers
 *	and its associated geometry.
 */

void
OPScrollView::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->h_scroller << this->v_scroller;
	storage << this->x_offset_ratio << this->y_offset_ratio;
	storage << this->x_offset << this->y_offset;
	storage << this->width << this->height;
	storage << OPStorageSync::mark;
}



/*
 *	Restore a Scroll View with its links to the scrollers
 *	and its associated geometry.
 */

void
OPScrollView::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->h_scroller >> this->v_scroller;
	storage >> this->x_offset_ratio >> this->y_offset_ratio;
	storage >> this->x_offset >> this->y_offset;
	storage >> this->width >> this->height;
	storage >> OPStorageSync::mark;
}



/*
 *	Prepare the scroll view for destruction. Give up both
 *	links to the scrollers.
 */

void
OPScrollView::Delete ()
{
	OP_METHOD (Delete, 0);
	
	OPScrollerRef zero = 0;
	
	this->SetHScroller (zero);
	this->SetVScroller (zero);
	
	super::Delete ();
}



/*
 *	Scroll the view's contents by the specified amount of
 *	points.
 *	
 *	NIY : Optimized redrawing !!!
 */

void
OPScrollView::ScrollBy (OPRelCoord dx, OPRelCoord dy)
{
	OP_METHOD (ScrollBy, 2);
	OP_ARG (1, OPRelCoord, dx);
	OP_ARG (2, OPRelCoord, dy);
	
	if ( (dx == 0)
	  && (dy == 0) ) {
		return;
	}
	
	this->x_offset += dx;
	this->y_offset += dy;
	
	OPViewRef child;
	
	if (this->GetFirstChild (child)) {
		
		while (child.IsValid ()) {
			child->ChangeOrigin (-dx, -dy);
			child->GetNextSibling (child);
		}
		
		//	Redraw the view. This could be done in a quite more efficient
		//	way : just redraw the missing parts.
		
		this->Display ();
	}
}



/*
 *	The scroll ratio changed. This will send notifications
 *	to the scrollers (if they are scrollers, what is not
 *	an obligation).
 */

void
OPScrollView::ScrollRatioChanged ()
{
	OP_METHOD (ScrollRatioChanged, 0);
	
	Real w = (this->width) ? (Real)(this->box.DX ()) / (Real)(width)
						   : (Real)(1.0);
	
	Real h = (this->height) ? (Real)(this->box.DY ()) / (Real)(height)
							: (Real)(1.0);
	
	if (w > 1.0) w = 1.0;
	if (h > 1.0) h = 1.0;
	
	//	Setup a message which will be sent to the scrollers. Since the type of
	//	the scrollers is not known for sure, we use some lose message passing.
	
	OPMessageRef msg = OP_NEW_FROM_ZONE (OPaC_temp_zone, OPMessage);
	
	if (this->h_scroller.IsValid ()) {
		
		//	Horizontal scroller will be notified of change and will
		//	eventually be redrawn.
		
		OPValueRef value = w;
		
		this->h_scroller->DefineData (OPScrollView::name_ratio, value);
//		this->h_scroller <- (msg-OPScrollView::do_display);
	}
	
	if (this->v_scroller.IsValid ()) {
		
		//	Vertical scroller will be notified of change and will
		//	eventually be redrawn.
		
		OPValueRef value = h;
		
		this->v_scroller->DefineData (OPScrollView::name_ratio, value);
//		this->v_scroller <- (msg-OPScrollView::do_display);
	}
}



/*
 *	Define the offset of the contents. The value 0.0 means
 *	no offset (left or top) and 1.0 maximum offset (right
 *	or bottom).
 */

void
OPScrollView::SetOffset (Real x, Real y)
{
	OP_METHOD (SetOffset, 2);
	OP_ARG (1, Real, x);
	OP_ARG (2, Real, y);
	
	if (x > 1.0) x = 1.0; else if (x < 0.0) x = 0.0;
	if (y > 1.0) y = 1.0; else if (y < 0.0) y = 0.0;
	
	this->x_offset_ratio = x;
	this->y_offset_ratio = y;
	
	OPRelCoord dx = this->width - this->box.DX ();	if (dx < 0) dx = 0;
	OPRelCoord dy = this->height - this->box.DY ();	if (dy < 0) dy = 0;
	
	OPRelCoord xo = (OPRelCoord)(Int32)(x * dx + 0.5) - this->x_offset;
	OPRelCoord yo = (OPRelCoord)(Int32)(y * dy + 0.5) - this->y_offset;
	
	if ( (xo != 0) || (yo != 0) ) {
		
		//	The position really changed on screen, Move every subview
		//	contained in the scroll view.
		
		this->ScrollBy (xo, yo);
	}
}



/*
 *	Define the coordinate of the top/left visible point.
 *	The origin of the subviews is set to -x, -y relative
 *	to the top/left corner of the scroll view.
 */

void
OPScrollView::SetTopLeft (OPRelCoord x, OPRelCoord y)
{
	OP_METHOD (SetTopLeft, 2);
	OP_ARG (1, OPRelCoord, x);
	OP_ARG (2, OPRelCoord, y);
	
	OPRelCoord dx = this->width - this->box.DX ();	if (dx < 0) dx = 0;
	OPRelCoord dy = this->height - this->box.DY ();	if (dy < 0) dy = 0;
	
	if (x < 0) x = 0; else if (x > dx) x = dx;
	if (y < 0) y = 0; else if (y > dy) y = dy;
	
	OPRelCoord xo = x - this->x_offset;
	OPRelCoord yo = y - this->y_offset;
	
	if ( (xo != 0) || (yo != 0) ) {
		
		this->x_offset_ratio = (dx) ? ((Real)(x) / (Real)(dx)) : 0.0;
		this->y_offset_ratio = (dy) ? ((Real)(y) / (Real)(dy)) : 0.0;
		
		//	The position really changed on screen, Move every subview
		//	contained in the scroll view.
		
		this->ScrollBy (xo, yo);
	}
}



/*
 *	Set the total size of the scroll view.
 */

void
OPScrollView::SetTotalSize (OPCoord width, OPCoord height)
{
	OP_METHOD (SetTotalSize, 2);
	OP_ARG (1, OPCoord, width);
	OP_ARG (2, OPCoord, height);
	
	this->width  = width;
	this->height = height;
	this->SetOffset (this->x_offset_ratio, this->y_offset_ratio);
	this->ScrollRatioChanged ();
}



/*
 *	Set the horizontal scroller. The scroller may be any
 *	object as long as it behaves as a scroller (that is,
 *	it sends reports to notify about scrolls and accepts
 *	SetVisibleRatio / Display messages).
 */

void
OPScrollView::SetHScroller (OPObjectIn scr)
{
	OP_METHOD (SetHScroller, 1);
	OP_ARG (1, OPObjectIn, scr);
	
	OPScrollViewRef self = this;
	
	if (this->h_scroller.IsValid ()) {
		this->h_scroller->DetachObserver (self, OPScrollView::do_scroll, SCROLL_HOR);
	}
	
	if (scr.IsValid ()) {
		scr->AttachObserver (self, OPScrollView::do_scroll, SCROLL_HOR);
	}
	
	this->h_scroller = scr;
}


/*
 *	Set the vertical scroller. See SetHScroller.
 */

void
OPScrollView::SetVScroller (OPObjectIn scr)
{
	OP_METHOD (SetVScroller, 1);
	OP_ARG (1, OPObjectIn, scr);
	
	OPScrollViewRef self = this;
	
	if (this->v_scroller.IsValid ()) {
		this->v_scroller->DetachObserver (self, OPScrollView::do_scroll, SCROLL_VER);
	}
	
	if (scr.IsValid ()) {
		scr->AttachObserver (self, OPScrollView::do_scroll, SCROLL_VER);
	}
	
	this->v_scroller = scr;
}


/*
 *	The scroller tells the scrollview to update its
 *	positions.
 */

void
OPScrollView::Scroll (OPObjectIn object, Card32 id)
{
	OP_METHOD (Scroll, 2);
	OP_ARG (1, OPObjectIn, object);
	OP_ARG (2, Card32, id);
	
	OPValueRef value;
	
	//	Ask the scroller, which sent the "Scroll" message, about the current
	//	value associated to it. This is done dynamically :-)
	
	if (object->QueryData (OPScrollView::name_value, value)) {
		
		Real v = value;
		Real x = (this->h_scroller == object) ? v : this->x_offset_ratio;
		Real y = (this->v_scroller == object) ? v : this->y_offset_ratio;
		
		this->SetOffset (x, y);
	}
}



/*
 *	Draw the subviews. This uses the standard drawing
 *	algorithm, but limits the clipping rectangle.
 */

void
OPScrollView::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	super::Draw (port, rect, what);
}



/*
 *	Resizing a scroll view does not notify the children
 *	about it, since they should not see any difference.
 */

void
OPScrollView::SetSize (OPCoord dx, OPCoord dy)
{
	OP_METHOD (SetSize, 2);
	OP_ARG (1, OPCoord, dx);
	OP_ARG (2, OPCoord, dy);
	
	OP_ASSERT (dx != 0);
	OP_ASSERT (dy != 0);
	
	this->box.SetSize (dx, dy);
	this->ScrollRatioChanged ();
}



/*
 *	Moving the scroll view also moves the subviews...
 */

void
OPScrollView::SetOrigin (OPCoord x, OPCoord y)
{
	OP_METHOD (SetOrigin, 2);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	
	if ( (this->box.X1 () == x)
	  && (this->box.Y1 () == y) ) {
		return;
	}
	
	this->box.SetOrigin (x, y);
}


/*
 *	Move the scroll view's origin.
 */

void
OPScrollView::ChangeOrigin (OPRelCoord dx, OPRelCoord dy)
{
	OP_METHOD (ChangeOrigin, 2);
	OP_ARG (1, OPRelCoord, dx);
	OP_ARG (2, OPRelCoord, dy);
	
	super::ChangeOrigin (dx, dy);
}



/*
 *	Change the scroll view's size.
 */

void
OPScrollView::ChangeSize (OPRelCoord dx, OPRelCoord dy)
{
	OP_METHOD (ChangeSize, 2);
	OP_ARG (1, OPRelCoord, dx);
	OP_ARG (2, OPRelCoord, dy);
	
	super::ChangeSize (dx, dy);
}



/*
 *	Just change the scroll view's geometry and don't move
 *	any child around the world.
 */

void
OPScrollView::ChangeGeometry (OPRectIn box)
{
	OP_METHOD (ChangeGeometry, 1);
	OP_ARG (1, OPRectIn, box);
	
	super::ChangeGeometry (box);
//	this->box = box;
}


/*
 *	It is not the scroll view which accepts to become the main
 *	window, but it will be one of its sub-views.
 */

Bool
OPScrollView::WouldAcceptMain () const
{
	OP_METHOD (WouldAcceptMain, 0);

	return (this->active_child.head.IsValid ())
	    && (this->active_child.head->WouldAcceptMain ());
}



/*
 *	Notify the scroll view that it has become the main responder.
 *	This cannot be accepted and will 
 */

void
OPScrollView::BecomeMain (Bool clicked)
{
	OP_METHOD (BecomeMain, 1);
	OP_ARG (1, Bool, clicked);
	OP_ASSERT (this->active_child.head.IsValid ());
	
	this->window->SetMainResponder (this->active_child.head, clicked);
}

void
OPScrollView::ResignMain ()
{
	OP_METHOD (ResignMain, 0);
	super::ResignMain ();
}


