/*
 *	string.cxx
 *
 *	This file implements the string class, used by the OPaC project.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/string.h>
#include <OPaC/new.h>
#include <OPaC/except.h>
#include <OPaC/pubmacro.h>
#include <OPaC/values.h>

#include <string.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPString);

/****************************************************************/

OP_PUBLISH_MODIFIER (OPString, "Value", GetString, SetString, "String value", OPObject, OPString);

/****************************************************************/

#define	BUFFER_RESERVE	40

/****************************************************************/

/*
 *	Construct an empty string. The default will be an ISO
 *	Latin-1 string. This may change...
 */

OPString::OPString ()
{
	OP_METHOD (OPString, 0);
	
	this->text.latin = 0;
	this->type       = TEXT_IS_LATIN;
	this->len        = 0;
	this->buffer_len = 0;
}



/*
 *	Destroy the string. Free any remaining memory.
 */

OPString::~OPString ()
{
	OP_METHOD (~OPString, 0);
	
	this->FreeMemory ();
}



/*
 *	Save the string to the archive. The string will be
 *	stored in its natural format.
 */

void
OPString::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	Card8  type = (Card8)(this->type);
	Card32 len  = (Card32)(this->len);
	
	storage << type << len;
	storage << OPStorageSync::mark;
	
	if (len) {
		
		switch (this->type) {
			
			case TEXT_IS_LATIN:
			case TEXT_IS_SOLID_LATIN:
				storage->StoreBinData (this->text.latin, len * sizeof (LatinChar));
				break;
			
			case TEXT_IS_UNICODE:
			case TEXT_IS_SOLID_UNICODE:
				storage->StoreBinData (this->text.unicode, len * sizeof (UnicodeChar));
				break;
		}
	}
}



/*
 *	Restore the string from the archive. The string will
 *	no longer be solid if it had been when archived.
 */

void
OPString::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	OPMemZone* zone = this->ReturnMemZone ();
	
	Card8  type = 0;
	Card32 len  = 0;
	
	storage >> type >> len;
	storage >> OPStorageSync::mark;
	
	if (len == 0) {
		
		//	If there is nothing to restore, do not restore anything. It is
		//	sufficient to return, since the string was empty before being
		//	restored. Or wasn't it ?
		
		OP_ASSERT (this->len == 0);
		return;
	}
	
	switch ((AnyTextType)(type)) {
		
		case TEXT_IS_LATIN:
		case TEXT_IS_SOLID_LATIN:
			this->text.latin = OP_ALLOC_ARRAY_FROM_ZONE (zone, LatinChar, len);
			this->type       = TEXT_IS_LATIN;
			this->len        = len;
			this->buffer_len = len;
			storage->RestoreBinData (this->text.latin, len * sizeof (LatinChar));
			break;
		
		case TEXT_IS_UNICODE:
		case TEXT_IS_SOLID_UNICODE:
			this->text.unicode = OP_ALLOC_ARRAY_FROM_ZONE (zone, UnicodeChar, len);
			this->type         = TEXT_IS_LATIN;
			this->len          = len;
			this->buffer_len   = len;
			storage->RestoreBinData (this->text.unicode, len * sizeof (UnicodeChar));
			break;
	}
}



/*
 *	Prepare the string for its destruction.
 */

void
OPString::Delete ()
{
	OP_METHOD (Delete, 0);
	super::Delete ();
}



/*
 *	Free the memory associated to the string. If it had
 *	been defined as "solid", don't free anything.
 */

void
OPString::FreeMemory ()
{
	OP_METHOD (FreeMemory, 0);
	
	if (this->buffer_len == 0) {
		return;
	}
	
	OPMemZone* zone = this->ReturnMemZone ();
	
	switch (this->type) {
		
		case TEXT_IS_LATIN:
			OP_FREE_FROM_ZONE (zone, this->text.latin);
			this->text.latin = 0;
			break;
		
		case TEXT_IS_UNICODE:
			OP_FREE_FROM_ZONE (zone, this->text.unicode);
			this->text.unicode = 0;
			break;
		
		case TEXT_IS_SOLID_LATIN:
			this->text.solid_latin = 0;
			break;
		
		case TEXT_IS_SOLID_UNICODE:
			this->text.solid_unicode = 0;
			break;
	}
	
	this->len        = 0;
	this->buffer_len = 0;
}



/*
 *	Allocate memory for a string. Allocation will take care
 *	of Unicode/Latin characters. The size is specified in
 *	characters.
 *	
 *	If successful, the internal buffer will be replaced by
 *	the allocated amount of memory. The other fields will
 *	be initialised accordingly.
 *
 *	This is only possible if the type is not "solid".
 */

void
OPString::AllocMemory (AnyTextType type, Count len)
{
	OP_METHOD (AllocMemory, 2);
	OP_ARG (1, AnyTextType, type);
	OP_ARG (2, Count, len);
	
	OP_ASSERT (len > 0);
	OP_ASSERT (type != TEXT_IS_SOLID_LATIN);
	OP_ASSERT (type != TEXT_IS_SOLID_UNICODE);
	
	//	If the size requested is smaller than the currently available size, but
	//	not significantly different, keep the previous buffer...
	
	if ( (this->buffer_len >= len)
	  && (this->buffer_len <= (len + BUFFER_RESERVE) ) ) {
		return;
	}
	
	OPMemZone* zone = this->ReturnMemZone ();
	void*      data = this->text.latin;
	
	if ( (this->type == TEXT_IS_SOLID_LATIN)
	  || (this->type == TEXT_IS_SOLID_UNICODE) ) {
		data = 0;
	}
	
	
	//	First allocate memory and only then free the previously used memory.
	//	This way, we are sure not to modify the string if an exception is
	//	thrown.
	
	switch (type) {
		
		case TEXT_IS_LATIN:
			this->text.latin = OP_ALLOC_ARRAY_FROM_ZONE (zone, LatinChar, len);
			this->len        = 0;
			this->buffer_len = len;
			this->type       = type;
			
			if (data) {
				OP_FREE_FROM_ZONE (zone, (LatinChar*)(data));
			}
			
			break;
		
		case TEXT_IS_UNICODE:
			this->text.unicode = OP_ALLOC_ARRAY_FROM_ZONE (zone, UnicodeChar, len);
			this->len          = 0;
			this->buffer_len   = len;
			this->type         = type;
			
			if (data) {
				OP_FREE_FROM_ZONE (zone, (UnicodeChar*)(data));
			}
			
			break;
		
		default:
			OP_ABORT ("Illegal type");
			break;
	}
}



/*
 *	This is very similar to the AllocMemory method, but it
 *	does not destroy the string's contents (if any).
 *
 *	It is not possible to shrink memory under the current
 *	string's length. If you try to do it, noting will be
 *	done.
 *
 *	It is possible to grow a "solid" string. This will
 *	copy it and it will therefore no longer be solid.
 */

void
OPString::ResizeMemory (Count len)
{
	OP_METHOD (ResizeMemory, 1);
	OP_ARG (1, Count, len);
	
	if (len <= this->len) {
		
		//	No problem... Trying to reduce memory under the current string size
		//	won't do anyting.
		
		return;
	}
	
	OPMemZone*   zone    = this->ReturnMemZone ();
	LatinChar*   latin   = this->text.latin;
	UnicodeChar* unicode = this->text.unicode;
	
	//	First allocate memory, then copy the string and finally free the old
	//	one.
	
	switch (this->type) {
		
		case TEXT_IS_LATIN:
		case TEXT_IS_SOLID_LATIN:
			this->text.latin = OP_ALLOC_ARRAY_FROM_ZONE (zone, LatinChar, len);
			this->buffer_len = len;
			
			if (latin) {
				
				//	Copy the string. This is independent of the string's solidity.
				
				for (Count i = 0; i < this->len; i++) {
					this->text.latin[i] = latin[i];
				}
				
				if (this->type == TEXT_IS_LATIN) {
					OP_FREE_FROM_ZONE (zone, latin);
				}
			}
			
			this->type = TEXT_IS_LATIN;
			break;
		
		case TEXT_IS_UNICODE:
			this->text.unicode = OP_ALLOC_ARRAY_FROM_ZONE (zone, UnicodeChar, len);
			this->buffer_len   = len;
			this->type         = type;
			
			if (unicode) {
				
				//	Copy the string. This is independent of the string's solidity.
				
				for (Count i = 0; i < this->len; i++) {
					this->text.unicode[i] = unicode[i];
				}
				
				if (this->type == TEXT_IS_UNICODE) {
					OP_FREE_FROM_ZONE (zone, unicode);
				}
			}
			
			this->type = TEXT_IS_UNICODE;
			break;
		
		default:
			OP_ABORT ("Illegal type");
			break;
	}
}



/*
 *	Define the string. It will be the copy of another
 *	string object.
 */

Bool
OPString::SetString (OPStringIn text)
{
	OP_METHOD (SetString, 1);
	OP_ARG (1, OPStringIn, text);
	return this->SetString (text, 0);
}


/*
 *	Return a quick access to the internal representation of
 *	the string.
 */

const LatinChar*
OPString::ReturnQuickLatin () const
{
	OP_METHOD (ReturnQuickLatin, 0);
	
	switch (this->type) {
		case TEXT_IS_LATIN:
		case TEXT_IS_SOLID_LATIN:
			return this->text.latin;
		default:
			break;
	}

	return 0;
}

const UnicodeChar*
OPString::ReturnQuickUnicode () const
{
	OP_METHOD (ReturnQuickUnicode, 0);
	
	switch (this->type) {
		case TEXT_IS_UNICODE:
		case TEXT_IS_SOLID_UNICODE:
			return this->text.unicode;
		default:
			break;
	}

	return 0;
}



/*
 *	Return the string, that is : self...
 */

Bool
OPString::GetString (OPStringOut text) const
{
	OP_METHOD (GetString, 1);
	OP_ARG (1, OPStringOut, text);
	
	OPStringRef self = (OPString*)(Address)(this);
	text = "";
	return text->SetString (self);
}

OPStringRef
OPString::ReturnString () const
{
	OP_METHOD (ReturnString, 0);
	
	OPStringRef self = (OPString*)(Address)(this);
	OPStringRef text = "";
	text->SetString (self);
	
	return text;
}



/*
 *	Define the string. It is stored as an ISO-Latin-1
 *	string.
 */

Bool
OPString::SetString (const LatinChar* text, Count srclen)
{
	OP_METHOD (SetString, 2);
	OP_ARG (1, const LatinChar*, text);
	OP_ARG (2, Count, srclen);
	
	//	Compute the length of the Latin string :
	
	const LatinChar* ptr = text;
	Count len = 0; while (*ptr++) len++;

	if ( (len > srclen)
	  && (srclen > 0) ) {
		len = srclen;
	}
	
	if (len == 0) {
		this->SetSolidString ("");
		return TRUE;
	}
	
	//	Allocate memory for the string :
	
	this->AllocMemory (TEXT_IS_LATIN, len + (BUFFER_RESERVE/2) + 1);
	
	this->len = len;
	LatinChar* dest = this->text.latin;
	while (len--) *dest++ = *text++;
	*dest = 0;
	
	return TRUE;
}



/*
 *	Define the string. It is stored as a Unicode string.
 */

Bool
OPString::SetString (const UnicodeChar* text, Count srclen)
{
	OP_METHOD (SetString, 2);
	OP_ARG (1, const UnicodeChar*, text);
	OP_ARG (2, Count, srclen);
	
	//	Compute the length of the Latin string :
	
	const UnicodeChar* ptr = text;
	Count len = 0; while (*ptr++) len++;
	
	if ( (len > srclen)
	  && (srclen > 0) ) {
		len = srclen;
	}
	
	if (len == 0) {
		return this->SetSolidString ("");
	}
	
	//	Allocate memory for the string :
	
	this->AllocMemory (TEXT_IS_UNICODE, len + (BUFFER_RESERVE/2) + 1);
	
	this->len = len;
	UnicodeChar* dest = this->text.unicode;
	while (len--) *dest++ = *text++;
	*dest = 0;
	
	return TRUE;
}



/*
 *	Define the string. It will be the copy of another
 *	string object.
 */

Bool
OPString::SetString (OPStringIn text, Count srclen)
{
	OP_METHOD (SetString, 2);
	OP_ARG (1, OPStringIn, text);
	OP_ARG (2, Count, srclen);
	
	const LatinChar*   src_l = text->text.latin;
	const UnicodeChar* src_u = text->text.unicode;
	
	Count        length = text->len;
	LatinChar*   dst_l  = 0;
	UnicodeChar* dst_u  = 0;
	
	if ( (length > srclen)
	  && (srclen > 0) ) {
		length = srclen;
	}
	
	//	Depending on the string type, we will have different work to do. Usually
	//	just copy the buffer...
	
	switch (text->type) {
		
		case TEXT_IS_LATIN:
			this->AllocMemory (TEXT_IS_LATIN, length + 1);
			this->len = length;
			dst_l = this->text.latin;
			while (length--) *dst_l++ = *src_l++;
			*dst_l = 0;
			break;
		
		case TEXT_IS_UNICODE:
			this->AllocMemory (TEXT_IS_UNICODE, length + 1);
			this->len = length;
			dst_u = this->text.unicode;
			while (length--) *dst_u++ = *src_u++;
			*dst_u = 0;
			break;
		
		case TEXT_IS_SOLID_LATIN:
			return this->SetSolidString (text->text.latin);
		
		case TEXT_IS_SOLID_UNICODE:
			return this->SetSolidString (text->text.unicode);
	}

	return TRUE;
}



/*
 *	Define the string as a "solid string" : it is constant
 *	and will stay forever at the specified address. We may
 *	just remember the copy -- no copy is involved.
 */

Bool
OPString::SetSolidString (const LatinChar* text)
{
	OP_METHOD (SetSolidString, 1);
	OP_ARG (1, const LatinChar*, text);
	
	this->FreeMemory ();
	this->type = TEXT_IS_SOLID_LATIN;
	this->text.solid_latin = text;
	
	return TRUE;
}



/*
 *	Define the string as a "solid string" : it is constant
 *	and will stay forever at the specified address. We may
 *	just remember the copy -- no copy is involved.
 */

Bool
OPString::SetSolidString (const UnicodeChar* text)
{
	OP_METHOD (SetSolidString, 1);
	OP_ARG (1, const UnicodeChar*, text);
	
	this->FreeMemory ();
	this->type = TEXT_IS_SOLID_UNICODE;
	this->text.solid_unicode = text;
	
	return TRUE;
}



/*
 *	Insert the specified zero-terminated string at the
 *	specified position...
 */

Bool
OPString::InsertString (Count pos, const LatinChar* text, Count srclen)
{
	OP_METHOD (InsertString, 3);
	OP_ARG (1, Count, pos);
	OP_ARG (2, const LatinChar*, text);
	OP_ARG (3, Count, srclen);
	
	if (this->len == 0) {
		return this->SetString (text, srclen);
	}
	
	Count length   = this->len;
	Count text_len = 0;
	const LatinChar* p = text;
	while (*p++) text_len++;
	
	if ( (text_len > srclen)
	  && (srclen > 0) ) {
		text_len = srclen;
	}
	
	if (pos > length) pos = length;
	
	//	Create a temporary buffer and reconstruct the string in three parts.
	//	Handle the Latin and Unicode cases separately :
		
	if ( (this->type == TEXT_IS_LATIN)
	  || (this->type == TEXT_IS_SOLID_LATIN) ) {
		
		Count      size   = sizeof (LatinChar) * (text_len + this->len + 1);
		LatinChar* buffer = (LatinChar*)(stack_alloc (size));
		LatinChar* dst    = buffer;
		LatinChar* src    = this->text.latin;
		
		while (pos--)		{ *dst++ = *src++; length--; }
		while (text_len--)	{ *dst++ = *text++; }
		while (length--)	{ *dst++ = *src++; }
		*dst = 0;

		return this->SetString (buffer);
		
	}

	Count        size   = sizeof (UnicodeChar) * (text_len + this->len + 1);
	UnicodeChar* buffer = (UnicodeChar*)(stack_alloc (size));
	UnicodeChar* dst    = buffer;
	UnicodeChar* src    = this->text.unicode;
	
	while (pos--)		{ *dst++ = *src++; length--; }
	while (text_len--)	{ *dst++ = OPString::LatinToUnicode (*text++); }
	while (length--)	{ *dst++ = *src++; }
	*dst = 0;
	
	return this->SetString (buffer);
}



/*
 *	Insert the specified zero-terminated string at the
 *	specified position...
 */

Bool
OPString::InsertString (Count pos, const UnicodeChar* text, Count srclen)
{
	OP_METHOD (InsertString, 3);
	OP_ARG (1, Count, pos);
	OP_ARG (2, const UnicodeChar*, text);
	OP_ARG (3, Count, srclen);
	
	if (this->len == 0) {
		return this->SetString (text, srclen);
	}
	
	Count length   = this->len;
	Count text_len = 0;
	const UnicodeChar* p = text;
	while (*p++) text_len++;
	
	if ( (text_len > srclen)
	  && (srclen > 0) ) {
		text_len = srclen;
	}
	
	if (pos > length) pos = length;
	
	//	Create a temporary buffer and reconstruct the string in three parts.
	//	Handle the Latin and Unicode cases separately :
		
	if ( (this->type == TEXT_IS_LATIN)
	  || (this->type == TEXT_IS_SOLID_LATIN) ) {
		
		Count      size   = sizeof (LatinChar) * (text_len + this->len + 1);
		LatinChar* buffer = (LatinChar*)(stack_alloc (size));
		LatinChar* dst    = buffer;
		LatinChar* src    = this->text.latin;
		
		while (pos--)		{ *dst++ = *src++; length--; }
		while (text_len--)	{ *dst++ = OPString::UnicodeToLatin (*text++); }
		while (length--)	{ *dst++ = *src++; }
		*dst = 0;

		return this->SetString (buffer);
		
	}
	
	Count        size   = sizeof (UnicodeChar) * (text_len + this->len + 1);
	UnicodeChar* buffer = (UnicodeChar*)(stack_alloc (size));
	UnicodeChar* dst    = buffer;
	UnicodeChar* src    = this->text.unicode;
	
	while (pos--)		{ *dst++ = *src++; length--; }
	while (text_len--)	{ *dst++ = *text++; }
	while (length--)	{ *dst++ = *src++; }
	*dst = 0;
	
	return this->SetString (buffer);
}



/*
 *	Insert the specified zero-terminated string at the
 *	specified position...
 */

Bool
OPString::InsertString (Count pos, OPStringIn text, Count srclen)
{
	OP_METHOD (InsertString, 3);
	OP_ARG (1, Count, pos);
	OP_ARG (2, OPStringIn, text);
	OP_ARG (3, Count, srclen);

	switch (text->type) {
		
		case TEXT_IS_LATIN:
			return this->InsertString (pos, text->text.latin, srclen);
		
		case TEXT_IS_SOLID_LATIN:
			return this->InsertString (pos, text->text.solid_latin, srclen);
			
		case TEXT_IS_UNICODE:
			return this->InsertString (pos, text->text.unicode, srclen);
			
		case TEXT_IS_SOLID_UNICODE:
			return this->InsertString (pos, text->text.solid_unicode, srclen);
	}
	
	return FALSE;
}



/*
 *	Append the specified string at the end of the current
 *	one.
 */

Bool
OPString::AppendString (const LatinChar* text, Count srclen)
{
	OP_METHOD (AppendString, 2);
	OP_ARG (1, const LatinChar*, text);
	OP_ARG (2, Count, srclen);
	
	//	This is not optimal -- just use the insertion method !
	
	return this->InsertString (this->len, text, srclen);
}



/*
 *	Append the specified string at the end of the current
 *	one.
 */

Bool
OPString::AppendString (const UnicodeChar* text, Count srclen)
{
	OP_METHOD (AppendString, 2);
	OP_ARG (1, const UnicodeChar*, text);
	OP_ARG (2, Count, srclen);
	
	//	This is not optimal -- just use the insertion method !
	
	return this->InsertString (this->len, text, srclen);
}



/*
 *	Append the specified string at the end of the current
 *	one.
 */

Bool
OPString::AppendString (OPStringIn text, Count srclen)
{
	OP_METHOD (AppendString, 2);
	OP_ARG (1, OPStringIn, text);
	OP_ARG (2, Count, srclen);
	
	//	This is not optimal -- just use the insertion method !
	
	if (srclen == 0) {
		srclen = text->len;
	}
	
	return this->InsertString (this->len, text, srclen);
}



/*
 *	Reset the string to an empty one...
 */

Bool
OPString::ClearString ()
{
	OP_METHOD (ClearString, 0);
	this->FreeMemory ();
	return TRUE;
}



/*
 *	Remove a piece of the string : the string will shrink.
 */

Bool
OPString::DeleteString (Count pos, Count num)
{
	OP_METHOD (DeleteString, 2);
	OP_ARG (1, Count, pos);
	OP_ARG (2, Count, num);
	
	if (pos >= this->len) {
		return TRUE;
	}
	
	if ( (pos == 0)
	  && (num >= this->len) ) {
		return this->ClearString ();
	}
	
	if ((pos + num) > this->len) {
		num = this->len - pos;
	}
	
	//	The position and the number of characters to delete are now both within
	//	valid limits :
	
	switch (this->type) {
		
		case TEXT_IS_LATIN:
			if ((pos + num) == this->len) {
				
				this->len = pos;
				
			} else {
				
				LatinChar* src = this->text.latin + pos + num;
				LatinChar* dst = this->text.latin + pos;
				Count      len = this->len - (pos + num);
				
				while (len--) { *dst++ = *src++; }
				
				this->len = this->len - num;
			}
			break;
		
		case TEXT_IS_SOLID_LATIN:
			this->SetString (this->text.solid_latin);
			this->DeleteString (pos, num);
			break;
			
		case TEXT_IS_UNICODE:
			if ((pos + num) == this->len) {
				
				this->len = pos;
				
			} else {
				
				UnicodeChar* src = this->text.unicode + pos + num;
				UnicodeChar* dst = this->text.unicode + pos;
				Count        len = this->len - (pos + num);
				
				while (len--) { *dst++ = *src++; }
				
				this->len = this->len - num;
			}
			break;
			
		case TEXT_IS_SOLID_UNICODE:
			this->SetString (this->text.solid_unicode);
			this->DeleteString (pos, num);
			break;
	}
	
	return TRUE;
}



/*
 *	Return number of characters stored in the current
 *	string...
 */

Count
OPString::CountCharacters () const
{
	return this->len;
}



/*
 *	Return TRUE if the default storage mode is Latin for
 *	this string.
 */

Bool
OPString::IsDefaultLatin () const
{
	OP_METHOD (IsDefaultLatin, 0);
	return ( (this->type == TEXT_IS_LATIN)
	      || (this->type == TEXT_IS_SOLID_LATIN) ) ? TRUE : FALSE;
}



/*
 *	Return TRUE if the default storage mode is Unicode for
 *	this string.
 */

Bool
OPString::IsDefaultUnicode () const
{
	OP_METHOD (IsDefaultLatin, 0);
	return ( (this->type == TEXT_IS_UNICODE)
	      || (this->type == TEXT_IS_SOLID_UNICODE) ) ? TRUE : FALSE;
}



/*
 *	If possible, return a pointer to the internal string
 *	representation in Latin. This method returns a null
 *	pointer if the data is not represented in Latin or
 *	if, for some reason, there is a problem.
 */

const LatinChar*
OPString::AccessLatin () const
{
	OP_METHOD (AccessLatin, 0);
	
	const LatinChar* text = 0;
	
	switch (this->type) {
		
		case TEXT_IS_LATIN:
			text = this->text.latin;
			break;
		
		case TEXT_IS_SOLID_LATIN:
			text = this->text.solid_latin;
			break;
		
		default:
			break;
	}

	return text;
}



/*
 *	If possible, return a pointer to the internal string
 *	representation in Unicode. This method returns a null
 *	pointer if the data is not represented in Unicode or
 *	if, for some reason, there is a problem.
 */

const UnicodeChar*
OPString::AccessUnicode () const
{
	OP_METHOD (AccessUnicode, 0);
	
	const UnicodeChar* text = 0;
	
	switch (this->type) {
		
		case TEXT_IS_UNICODE:
			text = this->text.unicode;
			break;
		
		case TEXT_IS_SOLID_UNICODE:
			text = this->text.solid_unicode;
			break;
		
		default:
			break;
	}

	return text;
}


/*
 *	Locate a character in the string. The search can
 *	be done forwards and backwards.
 */

Bool
OPString::Find (UnicodeChar what, CountOut pos, int dir) const
{
	OP_METHOD (Find, 3);
	OP_ARG (1, UnicodeChar, what);
	OP_ARG (2, CountOut, pos);
	OP_ARG (3, int, dir);
	
	if (dir >= 0) {
		Count index = 0;
		
		switch (this->type) {
			
			case TEXT_IS_LATIN:
			case TEXT_IS_SOLID_LATIN:
				while (index < this->len) {
					if (this->text.latin[index] == what) {
						pos = index;
						return TRUE;
					}
					index++;
				}
				break;
			
			case TEXT_IS_UNICODE:
			case TEXT_IS_SOLID_UNICODE:
				while (index < this->len) {
					if (this->text.unicode[index] == what) {
						pos = index;
						return TRUE;
					}
					index++;
				}
				break;
		}
		
	} else {
		
		Count index = this->len;
		
		switch (this->type) {
			
			case TEXT_IS_LATIN:
			case TEXT_IS_SOLID_LATIN:
				while (index--) {
					if (this->text.latin[index] == what) {
						pos = index;
						return TRUE;
					}
				}
				break;
			
			case TEXT_IS_UNICODE:
			case TEXT_IS_SOLID_UNICODE:
				while (index--) {
					if (this->text.unicode[index] == what) {
						pos = index;
						return TRUE;
					}
				}
				break;
		}
	}

	return FALSE;
}



Bool
OPString::Find (OPStringIn what, CountOut pos, int dir) const
{
	OP_METHOD (Find, 3);
	OP_ARG (1, UnicodeChar, what);
	OP_ARG (2, CountOut, pos);
	OP_ARG (3, int, dir);
	
	//	NIY: Unicode comparison.
	
	OP_ASSERT ((this->type == TEXT_IS_LATIN) || (this->type == TEXT_IS_SOLID_LATIN));
	OP_ASSERT ((what->type == TEXT_IS_LATIN) || (what->type == TEXT_IS_SOLID_LATIN));
	
	if (what->len == 0) {
		return TRUE;
	}
	
	if (what->len > this->len) {
		return FALSE;
	}
	
	const LatinChar* s1 = this->text.latin;
	const LatinChar* s2 = what->text.latin;
	Count l1 = this->len;
	Count l2 = what->len;
	
	if (dir >= 0) {
		while (pos+l2 <= l1) {
			if (strncmp (s1+pos, s2, l2) == 0) {
				return TRUE;
			}
			pos++;
		}
	} else {
		if (pos+l2 > l1) {
			pos = l1 - l2;
		}
		do {
			if (strncmp (s1+pos, s2, l2) == 0) {
				return TRUE;
			}
		} while (pos--);
	}
	
	return FALSE;
}

Bool
OPString::MakeLeft (Count num, OPStringOut text) const
{
	return this->MakeMiddle (0, num, text);
}

Bool
OPString::MakeMiddle (Count pos, Count num, OPStringOut text) const
{
	if (pos > this->len) pos = this->len;
	if (num > this->len-pos) num = this->len-pos;
	
	text = OP_NEW (OPString);
	
	if (num == 0) {
		return TRUE;
	}
	
	switch (this->type) {
		case TEXT_IS_LATIN:
		case TEXT_IS_SOLID_LATIN:
			text->SetString (this->text.latin+pos, num);
			break;
		case TEXT_IS_UNICODE:
		case TEXT_IS_SOLID_UNICODE:
			text->SetString (this->text.unicode+pos, num);
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

Bool
OPString::MakeRight (Count num, OPStringOut text) const
{
	if (num > this->len) num = this->len;
	return this->MakeMiddle (this->len-num, num, text);
}



/*
 *	Return the nth character from the specified position.
 *	If out of range, raise an exception.
 *
 *	Throws :
 *
 *	  - OPExcAccess (OUT_OF_RANGE) if the index is out of
 *		bounds.
 */

void
OPString::GetCharacter (Count pos, LatinCharOut character) const
{
	OP_METHOD (GetCharacter, 2);
	OP_ARG (1, Count, pos);
	OP_ARG (2, LatinCharOut, character);
	
	if (pos >= this->len) {
		THROW (OPExcAccess (OPExcAccess::OUT_OF_RANGE));
	}
	
	character = (LatinChar)(0);
	
	//	The conversion from Unicode to ISO Latin-1 is done in a simple way :
	//	just convert a single character and don't handle character pairs.
	
	switch (this->type) {
		
		case TEXT_IS_LATIN:
			character = this->text.latin[pos];
			break;
		
		case TEXT_IS_SOLID_LATIN:
			character = this->text.solid_latin[pos];
			break;
			
		case TEXT_IS_UNICODE:
			character = this->UnicodeToLatin (this->text.unicode[pos]);
			break;
			
		case TEXT_IS_SOLID_UNICODE:
			character = this->UnicodeToLatin (this->text.solid_unicode[pos]);
			break;
	}
}



/*
 *	Return the nth character from the specified position.
 *	If out of range, raise an exception.
 *
 *	Throws :
 *
 *	  - OPExcAccess (OUT_OF_RANGE) if the index is out of
 *		bounds.
 */

void
OPString::GetCharacter (Count pos, UnicodeCharOut character) const
{
	OP_METHOD (GetCharacter, 2);
	OP_ARG (1, Count, pos);
	OP_ARG (2, LatinCharOut, character);
	
	if (pos >= this->len) {
		THROW (OPExcAccess (OPExcAccess::OUT_OF_RANGE));
	}
	
	character = (UnicodeChar)(0);
	
	switch (this->type) {
		
		case TEXT_IS_LATIN:
			character = this->LatinToUnicode (this->text.latin[pos]);
			break;
		
		case TEXT_IS_SOLID_LATIN:
			character = this->LatinToUnicode (this->text.solid_latin[pos]);
			break;
			
		case TEXT_IS_UNICODE:
			character = this->text.unicode[pos];
			break;
			
		case TEXT_IS_SOLID_UNICODE:
			character = this->text.solid_unicode[pos];
			break;
	}
}



/*
 *	
 */

void
OPString::ConvertToUpper ()
{
	OP_ABORT ("NIY");
}



/*
 *	
 */

void
OPString::ConvertToLower ()
{
	OP_ABORT ("NIY");
}



/*
 *	
 */

void
OPString::ConvertToLatin ()
{
	switch (this->type) {
		case TEXT_IS_LATIN:
		case TEXT_IS_SOLID_LATIN:
			break;
		default:
			OP_ABORT ("NIY");
	}
}



/*
 *	
 */

void
OPString::ConvertToUnicode ()
{
	switch (this->type) {
		case TEXT_IS_UNICODE:
		case TEXT_IS_SOLID_UNICODE:
			break;
		default:
			OP_ABORT ("NIY");
	}
}



/*
 *	Extract a copy of the specified string. The buffer must
 *	take into account the additional terminal zero.
 */

Bool
OPString::GetStringCopy (LatinChar* buffer, Count max)
{
	OP_METHOD (GetStringCopy, 2);
	OP_ARG (1, LatinChar*, buffer);
	OP_ARG (2, Count, max);
	
	OP_ASSERT (buffer);
	
	if (max == 0) return TRUE;
	
	Count copy = (this->len < max) ? this->len : max - 1;
	
	const LatinChar*   src_l;
	const UnicodeChar* src_u;
	
	switch (this->type) {
		case TEXT_IS_LATIN:
		case TEXT_IS_SOLID_LATIN:
			src_l = this->text.solid_latin;
			while (copy--) *buffer++ = *src_l++;
			*buffer = 0;
			break;
			
		case TEXT_IS_UNICODE:
		case TEXT_IS_SOLID_UNICODE:
			src_u = this->text.solid_unicode;
			while (copy--) *buffer++ = OPString::UnicodeToLatin (*src_u++);
			*buffer = 0;
			break;
	}
	
	return (max > this->len) ? TRUE : FALSE;
}



/*
 *	
 */

Bool
OPString::GetStringCopy (UnicodeChar* buffer, Count max)
{
	OP_ABORT ("NIY");
	return FALSE;
}



/*
 *	Extract a substring. The number of characters which
 *	has been requested (num) does not take into account
 *	the final 0...
 *	
 *	The buffer has to be at least num+1 bytes long.
 *	Always return TRUE (?)
 */

Bool
OPString::ExtractStringCopy (LatinChar* buffer, Count pos, Count num)
{
	OP_METHOD (ExtractStringCopy, 3);
	OP_ARG (1, LatinChar*, buffer);
	OP_ARG (2, Count, pos);
	OP_ARG (3, Count, num);
	
	OP_ASSERT (buffer);
	
	buffer[0] = 0;
	
	if (num == 0) return TRUE;
	if (pos >= this->len) return TRUE;
	
	const LatinChar*   src_l;
	const UnicodeChar* src_u;
	
	switch (this->type) {
		case TEXT_IS_LATIN:
		case TEXT_IS_SOLID_LATIN:
			src_l = this->text.solid_latin + pos;
			while (num-- && (pos++ < this->len)) *buffer++ = *src_l++;
			*buffer = 0;
			break;
			
		case TEXT_IS_UNICODE:
		case TEXT_IS_SOLID_UNICODE:
			src_u = this->text.solid_unicode + pos;
			while (num-- && (pos++ < this->len)) *buffer++ = OPString::UnicodeToLatin (*src_u++);
			*buffer = 0;
			break;
	}
	
	return TRUE;
}



/*
 *	
 */

Bool
OPString::ExtractStringCopy (UnicodeChar* buffer, Count pos, Count num)
{
	OP_ABORT ("NIY");
	return FALSE;
}




/*
 *	Compute a hash index that can be used to store the string
 *	into a hash table.
 */

Count
OPString::ComputeHashValue (Count num) const
{
	OP_METHOD (ComputeHashValue, 1);
	OP_ARG (1, Count, num);
	OP_ASSERT (num);
	
	if (this->len == 0) {
		return 0;
	}
	
	OP_ASSERT ( (this->type == TEXT_IS_LATIN)
			 || (this->type == TEXT_IS_SOLID_LATIN) );
	
	Count hash = 0;
	Count seed = 0;
	Count temp = 0;
	
	const LatinChar* string = this->text.latin;
	
	while ( (temp = (Count)(*string++)) != 0 ) {
		seed += 13;
		hash += temp ^ seed;
	}
	
	return hash % num;
}



/*
 *	Compare contents and return an information which
 *	can be used for sorting purposes.
 */

int
OPString::CompareWith (OPObjectIn object) const
{
	OP_METHOD (CompareWith, 1);
	OP_ARG (1, OPObjectIn, object);
	
	OP_CLASS_CHECK (object, OPString);
	
	const OPStringRef string = object;
	
	//	NIY: Unicode comparison.
	
	OP_ASSERT ( (this->type == TEXT_IS_LATIN)
			 || (this->type == TEXT_IS_SOLID_LATIN) );
	
	OP_ASSERT ( (string->type == TEXT_IS_LATIN)
			 || (string->type == TEXT_IS_SOLID_LATIN) );
	
	if (string->len == 0) {
		return (this->len == 0) ? OP_COMP_THIS_SAME : OP_COMP_THIS_AFTER;
	}
	
	if (this->len == 0) {
		return OP_COMP_THIS_BEFORE;
	}
	
	const LatinChar* s1 = this->text.latin;
	const LatinChar* s2 = string->text.latin;
	Count l1 = this->len;
	Count l2 = string->len;
	
	while ( (l1-- != 0)
	     && (l2-- != 0) ) {
		
		if (*s1 == *s2) {
			s1++; s2++;
			continue;
		}
		
		return (*s1 < *s2) ? OP_COMP_THIS_BEFORE : OP_COMP_THIS_AFTER;
	}
	
	return OP_COMP_THIS_AFTER;
}



/*
 *	
 */

int
OPString::CompareWith (const LatinChar* s) const
{
	OP_ABORT ("NIY");
	return OP_COMP_THIS_AFTER;
}



/*
 *	
 */

int
OPString::CompareWith (const UnicodeChar* s) const
{
	OP_ABORT ("NIY");
	return OP_COMP_THIS_AFTER;
}



/*
 *	Convert an ISO Latin-1 character code into a Unicode
 *	character code. This is easy, since Latin-1 is the
 *	first page of Unicode...
 */

UnicodeChar
OPString::LatinToUnicode (LatinChar c)
{
	OP_MEMBER (LatinToUnicode, 1);
	OP_ARG (1, LatinChar, c);
	
	return (UnicodeChar)(c);
}



/*
 *	Convert a single Unicode character code into a Latin-1
 *	code. This does not handle character pairs...
 */

LatinChar
OPString::UnicodeToLatin (UnicodeChar c)
{
	OP_MEMBER (UnicodeToLatin, 1);
	OP_ARG (1, UnicodeChar, c);
	
	return (c < 256) ? (LatinChar)(c)
	                 : (LatinChar)(127);
}



/*
 *	Convert a Unicode character code into a Latin-1 code.
 *	A single glyph can be stored in several "characters",
 *	if there are some modifiers, for example.
 */

LatinChar
OPString::UnicodeToLatin (const UnicodeChar* c, CountInOut num)
{
	OP_MEMBER (UnicodeToLatin, 1);
	OP_ARG (1, UnicodeChar, c);
	
	if (num == 0) {
		return (LatinChar)(0);
	}
	
	//	NIY : complex multi-character glyphs are not handled properly !
	
	num = 1;
	
	return (c[0] < 256) ? (LatinChar)(c[0])
	                    : (LatinChar)(127);
}



/****************************************************************/

/*
 *	Atom creation from an anonymous object must handle the
 *	conversion for the known types.
 */

OPStringRef::OPStringRef (OPObjectIn x) : OPTextRef ()
{
	if (x.IsValid ()) {
		if (x->IsKindOf (OP_RUNTIME (OPAtom))) {
			OPAtomRef a = (OPAtom*)x.ReturnObject ();
			(*this) = a;
		} else if (x->IsKindOf (OP_RUNTIME (OPString))) {
			(*this) = (OPString*)x.ReturnObject ();
		} else if (x->IsKindOf (OP_RUNTIME (OPText))) {
			OPTextRef   b = (OPText*)x.ReturnObject ();
			OPStringRef s = b->ReturnString ();
			(*this) = s;
		} else if (x->IsKindOf (OP_RUNTIME (OPValue))) {
			OPValueRef  v = (OPValue*)x.ReturnObject ();
			OPStringRef s;
			v->ConvertToString (s);
			(*this) = s;
		} else {
			(*this) = (OPObject*)(0);
		}
	} else {
		this->_Set (OPObjectNull);
	}
}

OPStringRef::OPStringRef (OPStringIn x) : OPTextRef (x)
{
}

OPStringRef::OPStringRef (OPString* x) : OPTextRef (x)
{
}

OPStringRef&
OPStringRef::operator = (OPObjectIn x)
{
	if (x.IsValid ()) {
		if (x->IsKindOf (OP_RUNTIME (OPAtom))) {
			OPAtomRef a = (OPAtom*)x.ReturnObject ();
			(*this) = a;
		} else if (x->IsKindOf (OP_RUNTIME (OPString))) {
			(*this) = (OPString*)x.ReturnObject ();
		} else if (x->IsKindOf (OP_RUNTIME (OPText))) {
			OPTextRef   b = (OPText*)x.ReturnObject ();
			OPStringRef s = b->ReturnString ();
			(*this) = s;
		} else if (x->IsKindOf (OP_RUNTIME (OPValue))) {
			OPValueRef  v = (OPValue*)x.ReturnObject ();
			OPStringRef s;
			v->ConvertToString (s);
			(*this) = s;
		} else {
			(*this) = (OPObject*)(0);
		}
	} else {
		this->_Set (OPObjectNull);
	}
	return (*this);
}

OPStringRef&
OPStringRef::operator = (OPStringIn x)
{
	this->_Set (x);
	return (*this);
}

OPStringRef&
OPStringRef::operator = (OPString* x)
{
	this->_Set (x);
	return (*this);
}

/****************************************************************/

/*
 *	
 */

OPStringRef::OPStringRef (OPAtomIn string)
{
	OP_METHOD (OPStringRef, 1);
	OP_ARG (1, const LatinChar*, string);
	
	if (this->object == 0) {
		(*this) = OP_NEW (OPString);
	}
	
	const LatinChar* q_latin = string->ReturnQuickLatin ();
	if (q_latin) { (*this)->SetString (q_latin); return; }
	
	const UnicodeChar* q_unicode = string->ReturnQuickUnicode ();
	if (q_unicode) { (*this)->SetString (q_unicode); return; }
	
	(*this)->SetString (string->ReturnString ());
}



/*
 *	Initialise a string reference using a constant string
 *	pointer as initialiser.
 */

OPStringRef::OPStringRef (const LatinChar* string)
{
	OP_METHOD (OPStringRef, 1);
	OP_ARG (1, const LatinChar*, string);
	
	(*this) = OP_NEW (OPString);
	(*this)->SetString (string);
}



/*
 *	
 */

OPStringRef::OPStringRef (const UnicodeChar* string)
{
	OP_ABORT ("NIY");
}



/*
 *	
 */

OPStringRef&
OPStringRef::operator = (OPAtomIn string)
{
	OP_METHOD (operator=, 1);
	OP_ARG (1, const LatinChar*, string);
	
	if (this->object == 0) {
		(*this) = OP_NEW (OPString);
	}
	
	const LatinChar* q_latin = string->ReturnQuickLatin ();
	if (q_latin) { (*this)->SetString (q_latin); return *this; }
	
	const UnicodeChar* q_unicode = string->ReturnQuickUnicode ();
	if (q_unicode) { (*this)->SetString (q_unicode); return *this; }
	
	(*this)->SetString (string->ReturnString ());
	
	return *this;
}



/*
 *	
 */

OPStringRef&
OPStringRef::operator = (const LatinChar* string)
{
	OP_METHOD (operator=, 1);
	OP_ARG (1, const LatinChar*, string);
	
	if (this->object == 0) {
		(*this) = OP_NEW (OPString);
	}
	
	(*this)->SetString (string);
	
	return *this;
}



/*
 *	
 */

OPStringRef&
OPStringRef::operator = (const UnicodeChar* string)
{
	OP_ABORT ("NIY");
	return *this;
}
	

