/*
 *	builder.cxx
 *
 *	Manage the palette window, the main part of the OPaC dynamic
 *	interface builder.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "builder.h"
#include "connectline.h"
#include "buildedit.h"

#include <OPaC/archive.h>
#include <OPaC/boxview.h>
#include <OPaC/buildmisc.h>
#include <OPaC/button.h>
#include <OPaC/cellgrid.h>
#include <OPaC/dbscrlist.h>
#include <OPaC/dialog.h>
#include <OPaC/dragpal.h>
#include <OPaC/icon.h>
#include <OPaC/icostyle.h>
#include <OPaC/lookman.h>
#include <OPaC/menu.h>
#include <OPaC/scrgrid.h>
#include <OPaC/scroller.h>
#include <OPaC/selview.h>
#include <OPaC/selgrid.h>
#include <OPaC/slider.h>
#include <OPaC/stream.h>
#include <OPaC/switch.h>
#include <OPaC/tablist.h>
#include <OPaC/tabview.h>
#include <OPaC/textline.h>
#include <OPaC/textstyle.h>
#include <OPaC/texttag.h>
#include <OPaC/textcaption.h>
#include <OPaC/window.h>

#include <OPaC/event.h>
#include <OPaC/key.h>

#include <OPaC/datarep.h>
#include <OPaC/container.h>
#include <OPaC/new.h>

#include <OPaC/pubmacro.h>
#include <OPaC/targetmacro.h>

#include <OPaC/opacirt.h>

/****************************************************************/

APP_IMPLEMENT_CLASS (OPBuilder);

/****************************************************************/

OP_TARGET_ACTOR		(OPBuilder, SaveWindow);
OP_TARGET_OBSERVER	(OPBuilder, RemoveDialog);
OP_TARGET_ACTOR		(OPBuilder, CloneView);
OP_TARGET_ACTOR		(OPBuilder, OnClickPalette);
OP_TARGET_ACTOR		(OPBuilder, OnClickLock);

/****************************************************************/

OP_TARGET_ACTOR		(OPBuilder, NotifyMenuBegin);
OP_TARGET_OBSERVER	(OPBuilder, NotifyMenuEnd);
OP_TARGET_ACTOR		(OPBuilder, NotifyInspectSelection);
OP_TARGET_ACTOR		(OPBuilder, NotifyDeleteSelection);
OP_TARGET_ACTOR		(OPBuilder, NotifyConnectionBegin);
OP_TARGET_ACTOR		(OPBuilder, NotifyActiveWindowChanged);

OP_TARGET_CLASS (OPBuilder);
OP_TARGET_ARG (OPSelViewRef, sel);
OP_TARGET_CALL (NotifySelectionBegin (sel));
OP_TARGET_END ();

OP_TARGET_CLASS (OPBuilder);
OP_TARGET_ARG (OPSelViewRef, sel);
OP_TARGET_CALL (NotifySelectionEnd (sel));
OP_TARGET_END ();

OP_TARGET_CLASS (OPBuilder);
OP_TARGET_ARG (OPObjectRef, subst);
OP_TARGET_CALL (SubstituteTarget (subst));
OP_TARGET_END ();

OP_TARGET_CLASS (OPBuilder);
OP_TARGET_ARG (OPObjectRef, initiator);
OP_TARGET_ARG (OPObjectRef, object);
OP_TARGET_ARG (OPAtomRef, field);
OP_TARGET_CALL (EditSimpleObject (initiator, object, field));
OP_TARGET_END ();

/****************************************************************/

OP_PUBLISH_MODIFIER (OPBuilder, OP_BUILDER_OBJECT, GetEditObject, SetEditObject, "selected object", OPObject, OPObject);
OP_PUBLISH_MODIFIER (OPBuilder, OP_BUILDER_SELVIEW, GetSelView, SetSelView, "selected view", OPObject, OPSelView);

/****************************************************************/

OPBuilder::OPBuilder ()
{
	this->Initialise ();
}

OPBuilder::~OPBuilder ()
{
	this->palette = 0;
	this->w_dialog_no_edit = 0;
}


/*
 *	This method creates an editor for the specified object. If the
 *	object is simple, this is possible. Otherwise, no editor is
 *	created.
 */

void
OPBuilder::EditSimpleObject (OPObjectIn initiator, OPObjectIn object, OPAtomIn field)
{
	OPViewRef editor;
	
	if (object.IsInvalid ()) {
		return;
	}
	
	if (object->IsKindOf (OP_RUNTIME (OPTextTagBase))) {
		OPBuildEdit::EditTextProperty (initiator, object, "Caption", editor);
	}
	
	if (field.IsValid ()) {
		initiator->DefineData (field, editor);
	}
}

void
OPBuilder::Initialise ()
{
	OPApp::CurrentApp ()->AttachObserver (this, "ActiveWindowChanged", "NotifyActiveWindowChanged");
	
	this->ib_window = OP_NEW (OPWindow);
	this->ib_window->Initialise (20, 20, 320, 270, "Interface Builder");
	this->ib_window->SetStyle (OP_WIN_TITLE | OP_WIN_MOVE_OK | OP_WIN_CLOSE_OK);
	this->ib_window->SetFlagOkDrop (FALSE);
	
	this->palette = OP_NEW (OPView);
	this->palette->SetOrigin (0, 0);
	this->palette->SetSize (300, 220);
	this->palette->SetParent (this->ib_window);
	
	this->view_palette = OP_NEW (OPView);
	this->view_palette->SetOrigin (0, 0);
	this->view_palette->SetSize (300, 220);
	this->view_palette->SetParent (this->ib_window);
	
	this->icon_palette = OP_NEW (OPView);
	this->icon_palette->SetOrigin (0, 0);
	this->icon_palette->SetSize (300, 220);
	this->icon_palette->SetParent (this->ib_window);
	
	this->tab_view = OP_NEW (OPTabView);
	this->tab_list = OPTabListNull;
	
	OPListRef    list = OP_NEW (OPList);
	OPTextTagRef text;
	
	this->tab_view->SetParent (this->ib_window);
	this->tab_view->SetOrigin (8, 8);
	this->tab_view->SetSize (304, 250);
	this->tab_view->SetActiveTab (1);
	this->tab_view->SetTabNum (3);
	
	list->SetInsertMode (FALSE);
	
	text = OP_NEW (OPTextTag);	text->SetText ("Palette");		list->Insert (text);
	text = OP_NEW (OPTextTag);	text->SetText ("Views");		list->Insert (text);
	text = OP_NEW (OPTextTag);	text->SetText ("Icons");		list->Insert (text);
	
	this->tab_view->GetTabs (this->tab_list);
	this->tab_list->SetTabs (list);
	this->tab_view->GetViews (list);
	list->SetInsertMode (FALSE);
	
	this->SetupPalette ();
	this->palette->SetParent (OPViewNull);
	list->Insert (this->palette);
	
	this->SetupViewPalette ();
	this->view_palette->SetParent (OPViewNull);
	list->Insert (this->view_palette);
	
	this->SetupIconPalette ();
	this->icon_palette->SetParent (OPViewNull);
	list->Insert (this->icon_palette);
	
	this->SetupDialogNoEdit ();
	
/****************************************************************/
	
	this->ib_window->Open ();
	this->ib_window->Display ();
	this->ib_window->Hide ();
}



void
OPBuilder::SaveWindow ()
{
	if (this->active_window.IsInvalid ()) {
		return;
	}
	
	OPDataSlotRef slot = this->active_window->ReturnSlot ();
	
	if ( (slot.IsValid ())
	  && (slot->IsKindOf (OP_RUNTIME (OPDialog))) ) {
		
		OPDialogRef    dialog  = slot;
		OPArchiveRef   archive = OP_NEW (OPArchive);
		OPFileWriteRef file    = OP_NEW (OPFileWrite);
		OPAtomRef      name;
		
		dialog->GetName (name);
		
		file->OpenWrite (name->ReturnString ());
		archive->SetOutput (file);
		archive->SetDataSlot (slot);
		archive << this->active_window;
		file->Close ();
	}
}


/****************************************************************/

/*
 *	Setup the palette...
 */

void
OPBuilder::SetupPalette ()
{
	OPDragPaletteRef dragpal;
	OPButtonRef      button;
	OPTextLineRef    textline;
	OPTextTagRef     texttag;
	OPScrollerRef    scroller;
	OPSwitchRef      swibut;
	OPBoxViewRef     boxview;
	OPTextCaptionRef caption;
	OPSliderRef      slider;
	
	//	Create the button template.
	
	dragpal = OP_NEW (OPDragPalette);
	button  = OP_NEW (OPButton);
	
	button->Initialise (20, 20, 80, 24, "Button", this->palette);
	dragpal->Initialise (button);
	
	//	Create the text template.
	
	dragpal = OP_NEW (OPDragPalette);
	texttag = OP_NEW (OPTextTag);
	
	texttag->Initialise (20, 54, 80, 20, "Text", this->palette);
	dragpal->Initialise (texttag);
	
	//	Create the caption template
	
	dragpal = OP_NEW (OPDragPalette);
	caption = OP_NEW (OPTextCaption);
	
	caption->Initialise (20, 84, 80, 24, "Text", this->palette);
	dragpal->Initialise (caption);
	
	//	Create the box template.
	
	dragpal = OP_NEW (OPDragPalette);
	boxview = OP_NEW (OPBoxView);
	
	boxview->Initialise (20, 118, 80, 60, "Box", this->palette);
	dragpal->Initialise (boxview);
	
	//	Create the text line template.
	
	dragpal  = OP_NEW (OPDragPalette);
	textline = OP_NEW (OPTextLine);
	
	textline->SetParent (this->palette);
	textline->SetOrigin (130, 20);
	textline->SetSize (80, 24);
	
	dragpal->Initialise (textline);
	
	//	Create the radio switch button template (on and off)
	
	dragpal = OP_NEW (OPDragPalette);
	swibut  = OP_NEW (OPSwitch);
	
	swibut->Initialise (130, 63, 100, 20, "radio", this->palette);
	swibut->InitAsRadioButton ();
	swibut->TurnOff ();
	dragpal->Initialise (swibut);
	
	dragpal = OP_NEW (OPDragPalette);
	swibut  = OP_NEW (OPSwitch);
	
	swibut->Initialise (130, 85, 100, 20, "radio", this->palette);
	swibut->InitAsRadioButton ();
	swibut->TurnOn ();
	dragpal->Initialise (swibut);
	
	//	Create the check box button template (on and off)
	
	dragpal = OP_NEW (OPDragPalette);
	swibut  = OP_NEW (OPSwitch);
	
	swibut->Initialise (130, 132, 100, 20, "check box", this->palette);
	swibut->InitAsCheckBox ();
	swibut->TurnOff ();
	dragpal->Initialise (swibut);
	
	dragpal = OP_NEW (OPDragPalette);
	swibut  = OP_NEW (OPSwitch);
	
	swibut->Initialise (130, 154, 100, 20, "check box", this->palette);
	swibut->InitAsCheckBox ();
	swibut->TurnOn ();
	dragpal->Initialise (swibut);
	
	//	Create the horizontal & vertical scroller templates.
	
	dragpal  = OP_NEW (OPDragPalette);
	scroller = OP_NEW (OPScroller);
	
	scroller->SetParent (this->palette);
	scroller->Initialise (20, 188, 220, TRUE);
	dragpal->Initialise (scroller);
	
	dragpal  = OP_NEW (OPDragPalette);
	scroller = OP_NEW (OPScroller);
	
	scroller->SetParent (this->palette);
	scroller->Initialise (240, 75, 130, FALSE);
	dragpal->Initialise (scroller);
	
	dragpal = OP_NEW (OPDragPalette);
	slider  = OP_NEW (OPSlider);
	
	slider->SetParent (this->palette);
	slider->Initialise (220, 24, 40, TRUE);
	dragpal->Initialise (slider);
	
	dragpal = OP_NEW (OPDragPalette);
	slider  = OP_NEW (OPSlider);
	
	slider->SetParent (this->palette);
	slider->Initialise (265, 13, 40, FALSE);
	dragpal->Initialise (slider);
}

/****************************************************************/

/*
 *	Build a palette presenting the view classes, so that the user
 *	can instanciate one from its name...
 */

void
OPBuilder::SetupViewPalette ()
{
	Count max = OPClassMan::TotalCount ();
	Count index;
	
	OPDBScrollListRef list = OP_NEW (OPDBScrollList);
	
	list->SetParent (this->view_palette);
	list->SetOrigin (10, 10);
	list->SetSize (180, 170);
	list->SetupDataBase ("View classes");
	
	for (index = 0; index < max; index++) {
		
		OPRuntimeRef runtime;
		OPClassMan::GetClass (index, runtime);
		
		if (runtime->IsKindOf (OP_RUNTIME (OPView))) {
			OPAtomRef name = runtime->ReturnClassName ();
			list->Insert (name, name, runtime, name);
		}
	}
	
	list->Update ();
	this->view_list = list;
	
	this->view_drag = OP_NEW (OPDragPalette);
	this->view_drag->SetParent (this->view_palette);
	this->view_drag->SetOrigin (200, 10);
	this->view_drag->SetSize (80, 80);
	
	OPButtonRef button = OP_NEW (OPButton);
	
	button->Initialise (10, 190, 80, 24, "Clone", this->view_palette);
	button->AttachObserver (this, "OnClick", "CloneView");
}

void
OPBuilder::CloneView ()
{
	OPRuntimeRef runtime;
	
	if (this->view_list->GetSelectedData (runtime)) {
		
		OPViewRef view;
		
		while (this->view_drag->GetFirstChild (view)) {
			view->SetParent (OPViewNull);
			view->Delete ();
		}
		
		view = OP_DYNAMIC_NEW (runtime, OPView);
		
		if (view->IsKindOf (OP_RUNTIME (OPWindow))) {
			OPWindowRef window = view;
			window->SetOrigin (10, 10);
			window->SetSize (200, 200);
			window->Open ();
			window->Display ();
			window->ShowNormal ();
		} else {
			view->SetParent (this->view_drag);
			view->SetBox (this->view_drag->ReturnBox ());
			this->view_drag->Display ();
		}
	}
}

/****************************************************************/

void
OPBuilder::SetupIconPalette ()
{
	OPCellGridRef grid = OP_NEW (OPCellGrid);
	OPLookManRef  look = OPLookMan::Current ();
	
	grid->SetIsometricMode (FALSE);
	grid->SetArraySize (2, 1);
	
	//	Handle each public data, one after the other.
	
	for (Count index = 0; ; index++) {
		
		OPLookIcon icon = look->FindNthIcon (index);
		
		if (icon == OP_ICON_NONE) {
			break;
		}
		
		OPAtomRef        name;
		OPIconRef        view    = OP_NEW (OPIcon);
		OPTextTagRef     caption = OP_NEW (OPTextTag);
		OPDragPaletteRef drag    = OP_NEW (OPDragPalette);
		OPBoxViewRef     bview   = OP_NEW (OPBoxView);
		
		look->GetIconName (icon, name);
		
		if (name.IsInvalid ()) {
			name = "UNKNOWN ICON";
		}
		
		OPCoord dx;
		OPCoord dy;
		
		view->Initialise (icon);
		view->GetMinSize (dx, dy);
		drag->SetSize (dx+2, dy+2);
		view->SetParent (drag);
		view->SetSize (dx, dy);
		view->SetOrigin (1, 1);
		bview->SetDesiredMinSize (dx+2, dy+2);
		bview->EnableTitle (FALSE);
		bview->EnableFrame (FALSE);
		drag->SetParent (bview);
		
		caption->SetTitle (name);
		caption->GetMinSize (dx, dy);
		caption->SetSize (dx, dy);
		caption->SetMode (OP_TEXT_HPOS_LEFT | OP_TEXT_VPOS_CENTER);
		
		grid->InsertCellInColumn (bview, 0);
		grid->InsertCellInColumn (caption, 1);
	}
	
	grid->SetMargin (2);
	grid->UpdateGridGeometry ();
	
	OPScrollGridRef scroll = OP_NEW (OPScrollGrid);
	
	OPRect box (10, 10, 290, 210);
	
	scroll->SetParent (this->icon_palette);
	scroll->Initialise (grid, box);
}

/****************************************************************/

void
OPBuilder::SetupObjectBrowser (OPDataSlotIn slot)
{
	if (this->o_pane_window.IsInvalid ()) {
		this->o_pane_window = OP_NEW (OPWindow);
		this->o_pane_window->SetOrigin (20, 520);
		this->o_pane_window->SetSize (390, 110);
		this->o_pane_window->SetStyle (OP_WIN_TITLE | OP_WIN_MOVE_OK);
		this->o_pane_window->SetTitle ("Object Pane");
		
		OPTextTagRef tag;
		OPCtrlShape  shape;
		
		this->o_pane_palette_button = OP_NEW (OPButton);
		this->o_pane_palette_button->Initialise (5, 65, 80, 20, "Palette", this->o_pane_window);
		this->o_pane_palette_icon = OP_NEW (OPIcon);
		this->o_pane_palette_icon->Initialise (OP_ICON_GREEN_SLED_OFF);
		this->o_pane_palette_button->SetHiliteMode (TRUE);
		this->o_pane_palette_button->SetPressMode (FALSE);
		this->o_pane_palette_button->SetShortcut (this->o_pane_palette_icon);
		this->o_pane_palette_button->GetContents (tag);
		tag->SetMode (OP_TEXT_HPOS_LEFT | OP_TEXT_VPOS_CENTER);
		tag->SetLeftSpace (4);
		shape  = this->o_pane_palette_button->ReturnShapeStyle ();
		shape |= OP_CTRL_NOFRAME;
		this->o_pane_palette_button->SetShapeStyle (shape);
		this->o_pane_palette_button->AttachObserver (this, "OnClick", "OnClickPalette");
		
		this->o_pane_lock_button = OP_NEW (OPButton);
		this->o_pane_lock_button->Initialise (5, 85, 80, 20, "Lock", this->o_pane_window);
		this->o_pane_lock_icon = OP_NEW (OPIcon);
		this->o_pane_lock_icon->Initialise (OP_ICON_RED_SLED_OFF);
		this->o_pane_lock_button->SetHiliteMode (TRUE);
		this->o_pane_lock_button->SetPressMode (FALSE);
		this->o_pane_lock_button->SetShortcut (this->o_pane_lock_icon);
		this->o_pane_lock_button->GetContents (tag);
		tag->SetMode (OP_TEXT_HPOS_LEFT | OP_TEXT_VPOS_CENTER);
		tag->SetLeftSpace (4);
		shape  = this->o_pane_lock_button->ReturnShapeStyle ();
		shape |= OP_CTRL_NOFRAME;
		this->o_pane_lock_button->SetShapeStyle (shape);
		this->o_pane_lock_button->AttachObserver (this, "OnClick", "OnClickLock");
	}
	
	if (this->o_pane_provider.IsInvalid ()) {
		this->o_pane_provider = OP_NEW (OPDBProvider);
	}
	
	this->o_pane_provider->Initialise (slot->ReturnFieldBase ());
	
	if (this->o_pane_pane.IsInvalid ()) {
		
		OPTextTagRef tag;
		OPButtonRef button = OP_NEW (OPButton);
		button->Initialise (380-120, 5, 60, 24, "Save", this->o_pane_window);
		button->AttachObserver (this, "OnClick", "SaveWindow");
		button = OP_NEW (OPButton);
		button->Initialise (380-60, 5, 60, 24, "Open", this->o_pane_window);
//		button->AttachObserver (this, "OnClick", "OpenWindow");
		button = OP_NEW (OPButton);
		button->Initialise (110, 5, 145, 24, "Window", this->o_pane_window);
//		button->AttachObserver (this, "OnClick", "DeleteWindow");
		button->GetContents (tag);
		button->SetShapeStyle (OP_CTRL_NOFRAME);
		button->SetPressMode (FALSE);
		button->SetHiliteMode (TRUE);
		tag->SetLeftSpace (4);
		tag->SetMode (OP_TEXT_HPOS_LEFT | OP_TEXT_VPOS_CENTER);
		this->o_pane_window_button = button;
		
		OPSelGridRef grid;
		OPRect box;
		box.SetOrigin (110, 34);
		box.SetSize (270, 70);
		
		this->o_pane_pane = OP_NEW (OPObjectPane);
		this->o_pane_pane->SetParent (this->o_pane_window);
		this->o_pane_pane->Initialise (box, this->o_pane_provider, "OPSelGrid");
		this->o_pane_pane->GetCellGrid (grid);
		this->o_pane_pane->SetNameDX (80);
		
		grid->SetSelectOnClick (TRUE);
	}
	
	this->o_pane_pane->Update ();
	
	this->o_pane_window->Open ();
	this->o_pane_window->Display ();
	this->o_pane_window->ShowNormal ();
}


/****************************************************************/

void
OPBuilder::OnClickPalette ()
{
	if (this->o_pane_palette_icon->ReturnNormalIcon () == OP_ICON_GREEN_SLED_OFF) {
		this->o_pane_palette_icon->Initialise (OP_ICON_GREEN_SLED_ON);
		this->ib_window->ShowNormal ();
	} else {
		this->o_pane_palette_icon->Initialise (OP_ICON_GREEN_SLED_OFF);
		this->ib_window->Hide ();
	}
	this->o_pane_palette_icon->Display ();
}

void
OPBuilder::OnClickLock ()
{
	if (this->o_pane_lock_icon->ReturnNormalIcon () == OP_ICON_RED_SLED_OFF) {
		this->o_pane_lock_icon->Initialise (OP_ICON_RED_SLED_ON);
		OPApp::CurrentApp ()->SetBuilder (OPBuilderNull);
	} else {
		this->o_pane_lock_icon->Initialise (OP_ICON_RED_SLED_OFF);
		OPApp::CurrentApp ()->SetBuilder (this);
	}
	this->o_pane_lock_icon->Display ();
}

/****************************************************************/

/*
 *
 */

Bool
OPBuilder::SetEditObject (OPObjectIn object)
{
	this->edit = object;
	
	if ( (object.IsValid ())
	  && (object->IsKindOf (OP_RUNTIME (OPView))) ) {
		
		OPViewRef view = object;
		OPRect box = view->ReturnBox ();
		
		view->DragAndDrop (box, this, OP_DDP_CONNECT_SUBST);
	}
	
	return TRUE;
}

void
OPBuilder::SubstituteTarget (OPObjectIn subst)
{
	this->edit = subst;
}

Bool
OPBuilder::GetEditObject (OPObjectOut object) const
{
	object = this->edit;
	return TRUE;
}

Bool
OPBuilder::SetSelView (OPSelViewIn sel_view)
{
	if (sel_view == this->sel_view) {
		return TRUE;
	}
	if (this->sel_view.IsValid ()) {
//		this->sel_view->EndSelection ();
	}
	
	this->sel_view = sel_view;
	return TRUE;
}

Bool
OPBuilder::GetSelView (OPSelViewOut sel_view) const
{
	sel_view = this->sel_view;
	return TRUE;
}


void
OPBuilder::NotifySelectionBegin (OPSelViewIn sel)
{
	this->sel_view = sel;
}

void
OPBuilder::NotifySelectionEnd (OPSelViewIn sel)
{
	if (this->sel_view == sel) {
		this->sel_view = OPSelViewNull;
	}
}



/*
 *
 */

void
OPBuilder::NotifyInspectSelection ()
{
	OPCoord more_y  = 0;
	OPCoord delta_x = 320;

	OPBuildEditRef builder = OP_NEW (OPBuildEdit);
	builder->EditProperties (this->edit);
	return;
}

void
OPBuilder::RemoveDialog (OPObjectIn button, Card32 id)
{
	OPViewRef   view = button;
	OPWindowRef window;

	view->GetWindow (window);

	window->Hide ();
	window->Delete ();
}


/*
 *
 */

void
OPBuilder::NotifyDeleteSelection ()
{
	OPViewRef view = this->sel_view;
	
	if ( (view.IsInvalid ())
	  && (this->edit.IsValid ())
	  && (this->edit->IsKindOf (OP_RUNTIME (OPView))) ) {
		view = edit;
	}
	
	if (view.IsValid ()) {
		
		OPSelViewRef sel;
		OPWindowRef  window;
		OPViewRef    parent;
		
		OPRect box = view->ReturnBox ();
		
		view->GetParent (parent);
		view->GetWindow (window);
		
		if (view->IsKindOf (OP_RUNTIME (OPSelView))) {
			sel = view;
		} else if ( (parent.IsValid ())
				 && (parent->IsKindOf (OP_RUNTIME (OPSelView))) ) {
			sel = parent;
		}
		
		if (sel.IsValid ()) {
			view = sel->ReturnSelectedView ();
			sel->EndSelection ();
			sel->Delete ();
		}
		
		view->Delete ();
		window->Redraw (box);
	}
}





struct MenuDef
{
	OPLookIcon		icon;
	const char*		label;
	MenuDef*		submenu;
	const char*		action;
};


static OPMenuRef
CreateVerticalMenu (const MenuDef* def, OPObjectRef ref)
{
	OPMenuRef     menu = OP_NEW (OPMenu);
	OPCellGridRef grid = OP_NEW (OPCellGrid);
	
	const MenuDef* iter = def;
	Count num = 0;
	
	while (iter->label) {
		num++;
		iter++;
	}
	
	grid->SetArraySize (1, num);
	grid->SetMargin (2);
	grid->SetCellSize (100, 26);
	grid->UpdateGridGeometry ();
	
	iter = def;
	num  = 0;
	
	while (iter->label) {
		OPButtonRef but = OP_NEW (OPButton);
		OPIconRef   ico = OP_NEW (OPIcon);
		
		ico->Initialise (iter->icon);
		but->SetIcon (ico);
		but->SetText (iter->label, OP_TEXT_HPOS_LEFT | OP_TEXT_VPOS_CENTER, FALSE);
		but->SetSensitiveButtons (OP_EV_BUTTON_RIGHT | OP_EV_BUTTON_LEFT);
		grid->SetCell (0, num++, but);
		if (iter->submenu) {
			ico = OP_NEW (OPIcon);
			ico->Initialise (OP_ICON_SUBMENU_NORMAL);
			but->SetShortcut (ico);
			OPMenuRef submenu = CreateVerticalMenu (iter->submenu, ref);
			but->SetAttribute ("Submenu", submenu);
		}
		if (iter->action) {
			but->AttachObserver (ref, "OnClick", iter->action);
		}
		iter++;
	}
	
	menu->SetCellGrid (grid);
	
	return menu;
}



/*
 *
 */

void
OPBuilder::NotifyMenuBegin ()
{
	MenuDef desc[] =
	{
		{ OP_ICON_NONE, "Inspector...", 0, OP_BUILDER_DO_INSPECT },
		{ OP_ICON_NONE, "Delete", 0, OP_BUILDER_DO_DELETE },
		{ OP_ICON_NONE, 0, 0 }
	};
	
	OPViewRef view = this->sel_view;
	
	if ( (view.IsInvalid ())
	  && (this->edit.IsValid ())
	  && (this->edit->IsKindOf (OP_RUNTIME (OPView))) ) {
		view = edit;
	}
	
	if (view.IsValid ()) {
		
		OPMenuRef menu = CreateVerticalMenu (desc, this);
		
		menu->PopupAsPopup (view, 0);
		menu->AttachObserver (this, "ButtonReleased", "NotifyMenuEnd");
		
		OPEvent::HandOverToWindow (menu);
	}
}


/*
 *	Remove the menu after the user released the button.
 *	This is invoked automatically by the menu.
 */

void
OPBuilder::NotifyMenuEnd (OPObjectIn object, Card32 id)
{
	OPMenuRef menu = object;
	menu->HideMenu ();
	menu->Delete ();
}




void
OPBuilder::SetupDialogNoEdit ()
{
	OPWindowRef window = OP_NEW (OPWindow);
	
	window->SetOrigin (50, 50);
	window->SetSize (200, 100);
	window->SetStyle (OP_WIN_TITLE | OP_WIN_MOVE_OK);
	window->SetTitle ("Edit attributes");
	
	OPButtonRef  button  = OP_NEW (OPButton);
	OPTextTagRef buttext = OP_NEW (OPTextTag);
	OPTextTagRef msgtext = OP_NEW (OPTextTag);
	
	msgtext->SetParent (window);
	msgtext->SetOrigin (10, 20);
	msgtext->SetSize (180, 20);
	msgtext->SetText ("This object cannot be edited.");
	msgtext->SetMode (OP_TEXT_HPOS_LEFT | OP_TEXT_VPOS_TOP);
	
	buttext->SetText ("OK");
	button->SetParent (window);
	button->SetOrigin (10, 60);
	button->SetSize (80, 24);
	button->SetContents (buttext);
	button->AttachObserver (window, "OnClick", "Hide");

	this->w_dialog_no_edit = window;
	
	window->Open ();
	window->Display ();
	window->Hide ();
}

void
OPBuilder::ShowDialogNoEdit ()
{
	this->w_dialog_no_edit->ShowNormal ();
}


void
OPBuilder::NotifyConnectionBegin ()
{
	OPConnectLineRef line = OP_NEW (OPConnectLine);
	
	OPCoord ox = 0;
	OPCoord oy = 0;
	OPCoord wx = 0;
	OPCoord wy = 0;
	OPCoord dx = 0;
	OPCoord dy = 0;
	
	OPViewRef view = this->edit;
	OPWindowRef window;
	
	view->GetOrigin (ox, oy);
	view->GetSize (dx, dy);
	view->GetWindow (window);
	window->GetOrigin (wx, wy);
	
	line->SetParent (window);
	line->BeginDrag (this->edit, ox+wx+dx/2, oy+wy+dy/2);
	
	window->SetDraggingView (line);
}


void
OPBuilder::NotifyActiveWindowChanged ()
{
	OPWindowRef window = OPApp::CurrentApp ()->ReturnActiveWindow ();
	
	if (window.IsValid ()) {
		OPDataSlotRef slot = window->ReturnSlot ();
		if (slot.IsValid ()) {
			OPTextTagRef tag;
			OPStringRef title;
			this->active_window = window;
			this->SetupObjectBrowser (slot);
			this->o_pane_window_button->GetContents (tag);
			window->GetTitle (title);
			tag->SetText (title);
			this->o_pane_window->Display ();
		}
	}
}

